import math
from typing import List

from . import *


class Interaction(Identified):
    """The Interaction class provides more detailed description of how the
    Feature objects of a Component are intended to work together. For
    example, this class can be used to represent different forms of
    genetic regulation (e.g., transcriptional activation or
    repression), processes from the central dogma of biology
    (e.g. transcription and translation), and other basic molecular
    interactions (e.g., non-covalent binding or enzymatic
    phosphorylation). Each Interaction includes type properties that
    refer to descriptive ontology terms and hasParticipation
    properties that describe which Feature objects participate in
    which ways in the Interaction.

    """

    def __init__(self, types: List[str],
                 *, participations: List[Participation] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_INTERACTION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.types = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                 initial_value=types)
        self.participations = OwnedObject(self, SBOL_PARTICIPATIONS, 0, math.inf,
                                          initial_value=participations,
                                          type_constraint=Participation)


def build_interaction(identity: str, *, type_uri: str = SBOL_INTERACTION) -> SBOLObject:
    interaction_type = PYSBOL3_MISSING
    obj = Interaction([interaction_type], identity=identity, type_uri=type_uri)
    # Remove the dummy type
    obj._properties[SBOL_TYPE] = []
    return obj


Document.register_builder(SBOL_INTERACTION, build_interaction)




class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class SBOLError(Error):

    def __init__(self, message: str):
        self.message = message


class NamespaceError(Error):
    """Exception raised for namespace errors.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message: str):
        self.message = message



import unittest

import sbol3


class TestRange(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_creation(self):
        start = 1
        end = 10
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        self.assertIsNotNone(r)
        self.assertEqual(start, r.start)
        self.assertEqual(end, r.end)
        self.assertEqual(sbol3.SBOL_RANGE, r.type_uri)

    def test_invalid_create(self):
        start = 0
        end = 10
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        report = r.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))

        # end must be > 0, and end must be > start
        start = 1
        end = 0
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        report = r.validate()
        self.assertIsNotNone(report)
        self.assertEqual(2, len(report.errors))

        # end must be >= start
        start = 10
        end = 9
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        report = r.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))

        start = 7
        end = 7
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        self.assertEqual(start, r.start)
        self.assertEqual(end, r.end)
        report = r.validate()
        self.assertIsNotNone(report)
        self.assertEqual(0, len(report.errors))

    def test_keyword_args(self):
        # Test that all arguments, both required and optional, can be
        # specified by keyword
        seq_uri = 'https://example.com/pysbol3/seq1'
        start = 7
        end = 14
        range_uri = 'https://example.com/pysbol3/r1'
        r1 = sbol3.Range(sequence=seq_uri, end=end, start=start,
                         identity=range_uri)
        self.assertEqual(seq_uri, r1.sequence)
        self.assertEqual(start, r1.start)
        self.assertEqual(end, r1.end)
        self.assertEqual(range_uri, r1.identity)
        display_id = range_uri[range_uri.rindex('/') + 1:]
        self.assertEqual(display_id, r1.display_id)


class TestCut(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_creation(self):
        at = 1
        cut = sbol3.Cut(sbol3.PYSBOL3_MISSING, at)
        self.assertIsNotNone(cut)
        self.assertEqual(at, cut.at)
        at = 0
        cut = sbol3.Cut(sbol3.PYSBOL3_MISSING, at)
        self.assertIsNotNone(cut)
        self.assertEqual(at, cut.at)
        self.assertEqual(sbol3.SBOL_CUT, cut.type_uri)

    def test_invalid_create(self):
        # At must be >= 0
        at = -1
        cut = sbol3.Cut(sbol3.PYSBOL3_MISSING, at)
        report = cut.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))


class TestEntireSequence(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_creation(self):
        # EntireSequence has no properties, so there isn't much to test here
        es = sbol3.EntireSequence(sbol3.PYSBOL3_MISSING)
        self.assertIsNotNone(es)
        self.assertEqual(sbol3.SBOL_ENTIRE_SEQUENCE, es.type_uri)


if __name__ == '__main__':
    unittest.main()



import math
from typing import List, Union

from . import *


class VariableFeature(Identified):
    """As described above, the VariableComponent VariableFeature class
    specifies a variable and set of values that will replace one of
    the SubComponent Feature objects in the template of a
    CombinatorialDerivation. The variable is specified by the variable
    property, and the set of values is defined by the union of
    Component objects referred to by the variant, variantCollection,
    and variantDerivation properties.

    """

    def __init__(self, cardinality: str, variable: Union[Identified, str],
                 *, variants: List[Union[Identified, str]] = None,
                 variant_collections: List[Union[Identified, str]] = None,
                 variant_derivations: List[Union[Identified, str]] = None,
                 variant_measures: List[Union[Identified, str]] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_VARIABLE_FEATURE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        # Assign default values
        if cardinality is None:
            cardinality = SBOL_ZERO_OR_MORE
        if variable is None:
            variable = PYSBOL3_MISSING
        # Create properties
        self.cardinality = URIProperty(self, SBOL_CARDINALITY, 1, 1,
                                       initial_value=cardinality)
        self.variable = ReferencedObject(self, SBOL_VARIABLE, 1, 1,
                                         initial_value=variable)
        self.variants = ReferencedObject(self, SBOL_VARIANT, 0, math.inf,
                                         initial_value=variants)
        self.variant_collections = ReferencedObject(self, SBOL_VARIANT_COLLECTION,
                                                    0, math.inf,
                                                    initial_value=variant_collections)
        self.variant_derivations = ReferencedObject(self, SBOL_VARIANT_DERIVATION,
                                                    0, math.inf,
                                                    initial_value=variant_derivations)
        self.variant_measures = OwnedObject(self, SBOL_VARIANT_MEASURE,
                                            0, math.inf,
                                            initial_value=variant_measures)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        # Cardinality must be in the set of valid URIs
        valid_cardinalities = [SBOL_ONE, SBOL_ONE_OR_MORE, SBOL_ZERO_OR_MORE,
                               SBOL_ZERO_OR_ONE]
        if self.cardinality not in valid_cardinalities:
            message = f'{self.cardinality} is not a valid cardinality'
            report.addError(self.identity, None, message)
        if self.variable is None:
            message = 'VariableComponent.variable is required'
            report.addError(self.identity, None, message)
        return report


def build_variable_feature(identity: str, type_uri: str = SBOL_VARIABLE_FEATURE):
    """Used by Document to construct a VariableFeature when reading an SBOL file.
    """
    obj = VariableFeature(cardinality=PYSBOL3_MISSING, variable=PYSBOL3_MISSING,
                          identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_CARDINALITY] = []
    obj._properties[SBOL_VARIABLE] = []
    return obj


Document.register_builder(SBOL_VARIABLE_FEATURE, build_variable_feature)



import math
import posixpath
from typing import Union, List, Callable
from urllib.parse import urlparse

import rdflib

from . import *
from .utils import parse_class_name


class Identified(SBOLObject):
    """All SBOL-defined classes are directly or indirectly derived from
    the Identified abstract class. This inheritance means that all
    SBOL objects are uniquely identified using URIs that uniquely
    refer to these objects within an SBOL document or at locations on
    the World Wide Web.

    """

    def __init__(self, identity: str, type_uri: str,
                 *, name: str = None, description: str = None,
                 derived_from: List[str] = None, generated_by: List[str] = None,
                 measures: List[SBOLObject] = None) -> None:
        super().__init__(identity)
        self._display_id = TextProperty(self, SBOL_DISPLAY_ID, 0, 1)
        self.name = TextProperty(self, SBOL_NAME, 0, 1, initial_value=name)
        self.description = TextProperty(self, SBOL_DESCRIPTION, 0, 1,
                                        initial_value=description)
        self.derived_from = URIProperty(self, PROV_DERIVED_FROM, 0, math.inf,
                                        initial_value=derived_from)
        self.generated_by = URIProperty(self, PROV_GENERATED_BY, 0, math.inf,
                                        initial_value=generated_by)
        # The type_constraint for measures should really be Measure but
        # that's a circular dependency. Instead we make the type constraint
        # Identified to constrain it somewhat. Identified is the best we
        # can do since every other SBOL class requires Identified to be defined.
        self.measures = OwnedObject(self, SBOL_HAS_MEASURE, 0, math.inf,
                                    initial_value=measures,
                                    type_constraint=Identified)
        # Identity has been set by the SBOLObject constructor
        self._display_id = self._extract_display_id(self.identity)
        self._rdf_types = URIProperty(self, RDF_TYPE, 1, math.inf,
                                      initial_value=[type_uri])

    @staticmethod
    def _is_valid_display_id(display_id: str) -> bool:
        if display_id is None:
            return True
        # A display id [...] value MUST be composed of only alphanumeric
        # or underscore characters and MUST NOT begin with a digit.
        # (Section 6.1)
        #
        # Make sure everything other than underscores are alphanumeric, and
        # make sure the first character is not a digit.
        #
        # Note: This implies that '_' is not a valid displayId because `isalnum()`
        # will return false if the string is empty, which it will be after we filter
        # out the lone underscore.
        return display_id.replace('_', '').isalnum() and not display_id[0].isdigit()

    @staticmethod
    def _extract_display_id(identity: str) -> Union[None, str]:
        if not identity:
            return None
        parsed = urlparse(identity)
        if not (parsed.scheme and parsed.netloc and parsed.path):
            # if the identity is not a URL, we cannot extract a display id
            # and display id is optional in this case
            return None
        display_id = parsed.path.split('/')[-1]
        if Identified._is_valid_display_id(display_id):
            return display_id
        else:
            msg = f'"{display_id}" is not a valid displayId.'
            msg += ' A displayId MUST be composed of only alphanumeric'
            msg += ' or underscore characters and MUST NOT begin with a digit.'
            raise ValueError(msg)

    @property
    def type_uri(self):
        return self._rdf_types[0]

    def _validate_display_id(self, report: ValidationReport) -> None:
        if self.identity_is_url():
            if (self.display_id is not None and
                    Identified._is_valid_display_id(self.display_id) and
                    self.identity.endswith(self.display_id)):
                return
        else:
            if Identified._is_valid_display_id(self.display_id):
                return
        message = f'{self.display_id} is not a valid displayId for {self.identity}'
        report.addError(self.identity, None, message)

    def _update_identity(self, identity: str, display_id: str) -> None:
        """Updates the identity of an Identified when it is added to a
        parent object. SBOL compliant objects and URIs require updating
        whenever an owned object is added to a new parent.
        """
        if self._identity is not None:
            class_name = type(self).__name__
            msg = f'{class_name} already has identity {self.identity}'
            msg += ' and cannot be re-parented.'
            raise ValueError(msg)
        self._identity = identity
        self._display_id = display_id
        # Now cycle through any owned objects and update their identities
        for _, objects in self._owned_objects.items():
            for child in objects:
                if child.display_id:
                    new_display_id = child.display_id
                else:
                    # Generate a display id based on type and number
                    type_name = parse_class_name(child.type_uri)
                    counter_value = self.counter_value(type_name)
                    new_display_id = f'{type_name}{counter_value}'
                new_identity = posixpath.join(self.identity, new_display_id)
                child._update_identity(new_identity, new_display_id)

    def counter_value(self, type_name: str):
        result = 0
        for _, objects in self._owned_objects.items():
            for sibling in objects:
                if sibling.display_id and sibling.display_id.startswith(type_name):
                    counter_string = sibling.display_id[len(type_name):]
                    counter_int = int(counter_string)
                    if counter_int > result:
                        result = counter_int
        return result + 1

    @property
    def display_id(self):
        # display_id is a read only property
        return self._display_id

    @property
    def properties(self):
        return list(self._properties.keys())

    def clear_property(self, uri):
        """Clears the internal storage of a property based on the URI.
        This is for advanced usage only, and may cause inconsistent
        objects and/or graphs.

        USE WITH CARE.
        """
        # If the URI is not in _properties or _owned_objects
        # silently do nothing. This is for advanced usage, so
        # don't tell the user it wasn't there.
        if uri in self._properties:
            self._properties[uri] = []
        elif uri in self._owned_objects:
            self._owned_objects[uri] = []

    def _validate_properties(self, report: ValidationReport) -> None:
        """Call validate on all the properties. Pass the name of the
        property so the error message is more friendly.
        """
        for k, v in self.__dict__.items():
            if isinstance(v, Property):
                v.validate(k, report)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        if report is None:
            report = ValidationReport()
        # Do validations for Identified
        self._validate_display_id(report)
        self._validate_properties(report)
        # Validate all owned objects
        for object_list in self._owned_objects.values():
            for obj in object_list:
                obj.validate(report)
        return report

    def serialize(self, graph: rdflib.Graph):
        identity = rdflib.URIRef(self.identity)
        # graph.add((identity, rdflib.RDF.type, rdflib.URIRef(self.type_uri)))
        for prop, items in self._properties.items():
            if not items:
                continue
            rdf_prop = rdflib.URIRef(prop)
            for item in items:
                graph.add((identity, rdf_prop, item))
        for prop, items in self._owned_objects.items():
            if not items:
                continue
            rdf_prop = rdflib.URIRef(prop)
            for item in items:
                graph.add((identity, rdf_prop, rdflib.URIRef(item.identity)))
                item.serialize(graph)

    def accept(self, visitor: Callable[['Identified'], None]):
        """Implement the visitor pattern by invoking `visitor` on self
        and all child objects.
        """
        visitor(self)
        for object_list in self._owned_objects.values():
            for obj in object_list:
                obj.accept(visitor)



import math

from . import *
# Feature is not exported
from .feature import Feature
from .location import Location
from .typing import *


class SequenceFeature(Feature):

    def __init__(self, locations: List[Location],
                 *, roles: List[str] = None, orientation: str = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_SEQUENCE_FEATURE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         roles=roles, orientation=orientation, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.locations: oo_list = OwnedObject(self, SBOL_LOCATION,
                                              1, math.inf,
                                              type_constraint=Location,
                                              initial_value=locations)


def build_sequence_feature(identity: str,
                           *, type_uri: str = SBOL_SEQUENCE_FEATURE) -> SBOLObject:
    obj = SequenceFeature([EntireSequence(PYSBOL3_MISSING)],
                          identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._owned_objects[SBOL_LOCATION] = []
    return obj


Document.register_builder(SBOL_SEQUENCE_FEATURE, build_sequence_feature)



import posixpath
import unittest
import sbol3


class TestOwnedObject(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_identity_append(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        con1 = sbol3.Constraint(sbol3.SBOL_REPLACES,
                                'http://example.com/fake1',
                                'http://example.com/fake2')
        # The constraint's identity and display_id will be overwritten as
        # part of the append operation. SBOL Compliant URIs (identities)
        # use the class of the object and a counter to generate the
        # display_id.
        expected2 = posixpath.join(comp.identity, 'Constraint1')
        # Append should cause the constraint's identity to change
        comp.constraints.append(con1)
        self.assertEqual(expected2, con1.identity)

    def test_list_property_update_identity(self):
        # This test uses assignment instead of appending
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        con1 = sbol3.Constraint(sbol3.SBOL_REPLACES,
                                'http://example.com/fake1',
                                'http://example.com/fake2')
        expected2 = posixpath.join(comp.identity, 'Constraint1')
        # Setting to list should cause the constraint's identity to change
        comp.constraints = [con1]
        self.assertEqual(expected2, con1.identity)

    def test_singleton_property_update_identity(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        tl = sbol3.CustomTopLevel('foo', 'http://synbio.bbn.com/opil#MeasurementValue')
        tl.measure = sbol3.OwnedObject(tl, 'http://synbio.bbn.com/opil#measure', 0, 1)
        m = sbol3.Measure(10, 'liters')
        tl.measure = m
        expected = posixpath.join(tl.identity, 'Measure1')
        self.assertIsNotNone(tl.measure.identity)
        self.assertEqual(expected, tl.measure.identity)

    def test_add_multiple_children(self):
        # Test that the the display_id and identity are overwritten
        # properly when adding multiple child entities.
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        comp.constraints.append(sbol3.Constraint(sbol3.SBOL_REPLACES,
                                                 'http://example.com/fake1',
                                                 'http://example.com/fake2'))
        expected1 = posixpath.join(comp.identity, 'Constraint1')
        self.assertEqual(expected1, comp.constraints[0].identity)
        comp.constraints.append(sbol3.Constraint(sbol3.SBOL_REPLACES,
                                                 'http://example.com/fake1',
                                                 'http://example.com/fake2'))
        expected2 = posixpath.join(comp.identity, 'Constraint2')
        self.assertEqual(expected2, comp.constraints[1].identity)

    def test_identity_conflict2(self):
        # Test that the same display id will cause a validation
        # error when an item with the same display id is appended
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        constraints = [sbol3.Constraint(sbol3.SBOL_REPLACES,
                                        'http://example.com/fake1',
                                        'http://example.com/fake2'),
                       sbol3.Constraint(sbol3.SBOL_REPLACES,
                                        'http://example.com/fake1',
                                        'http://example.com/fake2')]
        comp.constraints = constraints
        expected1 = posixpath.join(comp.identity, 'Constraint1')
        self.assertEqual(expected1, comp.constraints[0].identity)
        expected2 = posixpath.join(comp.identity, 'Constraint2')
        self.assertEqual(expected2, comp.constraints[1].identity)

    def test_cascade_identity(self):
        # Test that updating identity of an owned object cascades
        # to child owned objects
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        seq = sbol3.Sequence('seq1')
        loc = sbol3.EntireSequence(seq)
        seq_feature = sbol3.SequenceFeature([loc])
        c1.features.append(seq_feature)
        self.assertIsNotNone(seq_feature.identity)
        # identity should cascade down to the location after it
        # is set on the sequence feature
        self.assertIsNotNone(loc.identity)

    def test_overwrite_identity(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        seq = sbol3.Sequence('seq1')
        loc = sbol3.EntireSequence(seq)
        seq_feature = sbol3.SequenceFeature([loc])
        c1.features.append(seq_feature)
        self.assertIsNotNone(seq_feature.identity)
        # identity should cascade down to the location after it
        # is set on the sequence feature
        self.assertIsNotNone(loc.identity)
        old_sf_identity = seq_feature.identity
        old_loc_identity = loc.identity
        c2 = sbol3.Component('c2', sbol3.SBO_DNA)
        # Try adding the same object to a different parent
        # This should cause an error because the object is
        # still parented by c1.
        # See https://github.com/SynBioDex/pySBOL3/issues/178
        with self.assertRaises(ValueError):
            c2.features.append(seq_feature)
        self.assertEqual(old_loc_identity, loc.identity)
        self.assertEqual(old_sf_identity, seq_feature.identity)

    def test_type_constraint(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c = sbol3.Component('foo', sbol3.SBO_DNA)
        with self.assertRaises(TypeError):
            c.features = [sbol3.Range('https://example.com/fake', 1, 2)]
        self.assertEqual(c.features.type_constraint, sbol3.Feature)

    def test_set_document(self):
        # Ensure that document is set on child objects
        # See https://github.com/SynBioDex/pySBOL3/issues/176
        sbol3.set_namespace('https://bioprotocols.org/paml/primitives/')
        doc = sbol3.Document()
        c = sbol3.Component("scratch", sbol3.SBO_DNA)
        doc.add(c)
        lsc = sbol3.LocalSubComponent([sbol3.SBO_DNA])
        self.assertIsNone(lsc.document)
        c.features.append(lsc)
        self.assertIsNotNone(lsc.document)
        interaction = sbol3.Interaction([sbol3.SBO_DEGRADATION])
        c.interactions.append(interaction)
        self.assertEqual(doc, interaction.document)
        p = sbol3.Participation([sbol3.SBO_REACTANT], lsc)
        interaction.participations.append(p)
        self.assertEqual(doc, p.document)
        resolved_lsc = p.participant.lookup()
        self.assertEqual(lsc, resolved_lsc)
        self.assertEqual(lsc.identity, resolved_lsc.identity)

    # TODO: Write tests for adding via a slice
    #       comp.constraints[0:1] = sbol3.Constraint('foo')


if __name__ == '__main__':
    unittest.main()



import math
from typing import List

from . import *


class Collection(TopLevel):
    """The Collection class is a class that groups together a set of
    TopLevel objects that have something in common.

    Some examples of Collection objects:

    * Results of a query to find all Component objects in a repository
      that function as promoters.
    * A set of Component objects representing a library of genetic
      logic gates.
    * A "parts list" for Component with a complex design, containing
      both that component and all of the Component, Sequence, and
      Model objects used to provide its full specification.

    """

    def __init__(self, identity: str,
                 *, members: List[str] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_COLLECTION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.members = ReferencedObject(self, SBOL_MEMBER, 0, math.inf,
                                        initial_value=members)


class Namespace(Collection):
    """The Namespace class is a subclass of Collection and is used to
    define member entities that share the same URI prefix. Namely, all
    linked objects MUST have a URI prefix matching the URI of the
    Namespace object.

    """

    def __init__(self, identity: str,
                 *, members: List[str] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_NAMESPACE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         members=members,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


class Experiment(Collection):
    """The purpose of the Experiment class is to aggregate
    ExperimentalData objects for subsequent analysis, usually in
    accordance with an experimental design. Namely, the member
    properties of an Experiment MUST refer to ExperimentalData
    objects.

    """

    def __init__(self, identity: str,
                 *, members: List[str] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_EXPERIMENT) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         members=members,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


Document.register_builder(SBOL_COLLECTION, Collection)
Document.register_builder(SBOL_NAMESPACE, Namespace)
Document.register_builder(SBOL_EXPERIMENT, Experiment)



from typing import List

from . import *


class Model(TopLevel):
    """The purpose of the Model class is to serve as a placeholder for an
    external computational model and provide additional meta-data to
    enable better reasoning about the contents of this model. In this
    way, there is minimal duplication of standardization efforts and
    users of SBOL can elaborate descriptions of Component function in
    the language of their choice.

    """

    def __init__(self, identity: str, source: str, language: str,
                 framework: str,
                 *, attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_MODEL) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.source = URIProperty(self, SBOL_SOURCE, 1, 1,
                                  initial_value=source)
        self.language = URIProperty(self, SBOL_LANGUAGE, 1, 1,
                                    initial_value=language)
        self.framework = URIProperty(self, SBOL_FRAMEWORK, 1, 1,
                                     initial_value=framework)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        # The source property is REQUIRED and MUST contain a URI reference (Section 6.8)
        if not self.source:
            msg = f'Model {self.identity} does not have a source'
            report.addError(self.identity, None, msg)
        # The language property is REQUIRED and MUST contain a URI (Section 6.8)
        if not self.language:
            msg = f'Model {self.identity} does not have a language'
            report.addError(self.identity, None, msg)
        # The framework property is REQUIRED and MUST contain a URI (Section 6.8)
        if not self.framework:
            msg = f'Model {self.identity} does not have a framework'
            report.addError(self.identity, None, msg)
        return report


def build_model(identity: str, type_uri: str = SBOL_MODEL):
    """Used by Document to construct a Range when reading an SBOL file.
    """
    obj = Model(identity=identity, source=PYSBOL3_MISSING,
                language=PYSBOL3_MISSING, framework=PYSBOL3_MISSING,
                type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SOURCE] = []
    obj._properties[SBOL_LANGUAGE] = []
    obj._properties[SBOL_FRAMEWORK] = []
    return obj


Document.register_builder(SBOL_MODEL, build_model)



from typing import Sequence


class ValidationError:
    """A ValidationError is a violation of the SBOL specification.
    """

    def __init__(self, object_id, rule_id, message):
        self.rule_id = rule_id
        self.message = message
        self.object_id = object_id

    def __str__(self):
        result = f'{self.rule_id}: {self.message}'
        if self.object_id is not None:
            result = f'{self.object_id} {result}'
        return result


class ValidationWarning:
    """A ValidationWarning is a violation of an SBOL best practice.
    """

    def __init__(self, object_id, rule_id, message):
        self.rule_id = rule_id
        self.message = message
        self.object_id = object_id


class ValidationReport:

    def __init__(self):
        self._errors = []
        self._warnings = []

    def __len__(self):
        return len(self._errors) + len(self._warnings)

    @property
    def warnings(self) -> Sequence[ValidationWarning]:
        return tuple(self._warnings)

    @property
    def errors(self) -> Sequence[ValidationError]:
        return tuple(self._errors)

    def addError(self, object_id, rule_id, message):
        self._errors.append(ValidationError(object_id, rule_id, message))

    def addWarning(self, object_id, rule_id, message):
        self._warnings.append(ValidationWarning(object_id, rule_id, message))



import unittest

import sbol3


class TestExperimentalData(unittest.TestCase):

    # This class could use some more tests. Not sure what else to put
    # here. There are not yet any examples in the SBOLTestSuite. And
    # the class doesn't have any properties of its own.

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        display_id = 'exp_data'
        exp_data = sbol3.ExperimentalData(display_id)
        self.assertIsNotNone(exp_data)
        self.assertEqual(display_id, exp_data.display_id)
        self.assertTrue(hasattr(exp_data, 'attachments'))
        self.assertEqual(sbol3.SBOL_EXPERIMENTAL_DATA, exp_data.type_uri)

    def test_create2(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        display_id = 'exp_data'
        c1 = sbol3.Component('c1', [sbol3.SBO_DNA])
        exp_data = sbol3.ExperimentalData(display_id, attachments=[c1])
        self.assertIsNotNone(exp_data)
        self.assertEqual(display_id, exp_data.display_id)
        self.assertTrue(hasattr(exp_data, 'attachments'))
        self.assertEqual(sbol3.SBOL_EXPERIMENTAL_DATA, exp_data.type_uri)
        self.assertCountEqual([c1.identity], exp_data.attachments)


if __name__ == '__main__':
    unittest.main()



from typing import Union, List

from . import *


class Implementation(TopLevel):
    """An Implementation represents a realized instance of a Component,
    such a sample of DNA resulting from fabricating a genetic design
    or an aliquot of a specified reagent. Importantly, an
    Implementation can be associated with a laboratory sample that was
    already built, or that is planned to be built in the future. An
    Implementation can also represent virtual and simulated
    instances. An Implementation may be linked back to its original
    design using the prov:wasDerivedFrom property inherited from the
    Identified superclass. An Implementation may also link to a
    Component that specifies its realized structure and/or
    function.

    """

    def __init__(self, identity: str,
                 *, built: Union[Component, str] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None, generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_IMPLEMENTATION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.built = ReferencedObject(self, SBOL_BUILT, 0, 1,
                                      initial_value=built)


Document.register_builder(SBOL_IMPLEMENTATION, Implementation)



import unittest

import sbol3


class TestSequence(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        self.assertEqual(display_id, seq.display_id)
        self.assertIsNone(seq.elements)
        self.assertIsNone(seq.encoding)

    def test_invalid(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        seq.elements = 'actg'
        report = seq.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))

    def test_valid(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        seq.elements = 'actg'
        seq.encoding = sbol3.IUPAC_DNA_ENCODING
        # Should not raise a ValidationError
        report = seq.validate()
        self.assertEqual(0, len(report))

    def test_full_constructor(self):
        identity = 'https://github.com/synbiodex/pysbol3/s1'
        elements = 'GCAT'
        encoding = sbol3.IUPAC_DNA_ENCODING
        attachments = ['https://github.com/synbiodex/pysbol3/attachment1']
        name = None
        description = None
        derived_from = ['https://github.com/synbiodex/pysbol3/parent1']
        generated_by = ['https://github.com/synbiodex/pysbol3/tool1']
        m1 = sbol3.Measure(value=2.3, unit='meter')
        measures = [m1]
        s1 = sbol3.Sequence(identity=identity,
                            elements=elements,
                            encoding=encoding,
                            attachments=attachments,
                            name=name,
                            description=description,
                            derived_from=derived_from,
                            generated_by=generated_by,
                            measures=measures)
        self.assertEqual(identity, s1.identity)
        self.assertEqual(elements, s1.elements)
        self.assertEqual(encoding, s1.encoding)
        self.assertEqual(attachments, s1.attachments)
        self.assertEqual(name, s1.name)
        self.assertEqual(description, s1.description)
        self.assertEqual(derived_from, s1.derived_from)
        self.assertEqual(generated_by, s1.generated_by)
        self.assertEqual(measures, s1.measures)

    def test_invalid_encoding(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        seq.elements = 'actg'
        # This is an encoding from SBOL 3.0. It is no longer
        # valid as of 3.0.1/3.1.
        seq.encoding = 'http://sbols.org/v3#iupacNucleicAcid'
        # We expect 1 warning for the encoding that is not in the
        # recommended set.
        report = seq.validate()
        self.assertEqual(1, len(report))
        self.assertEqual(1, len(report.warnings))

    def test_initial_value(self):
        # See https://github.com/SynBioDex/pySBOL3/issues/208
        identity = 'https://github.com/synbiodex/pysbol3/s1'
        elements = ''
        # encoding = sbol3.IUPAC_DNA_ENCODING
        s1 = sbol3.Sequence(identity=identity,
                            elements=elements)
        self.assertEqual(identity, s1.identity)
        self.assertEqual(elements, s1.elements)
        # self.assertEqual(encoding, s1.encoding)


if __name__ == '__main__':
    unittest.main()



import math
import posixpath
import unittest
from collections import Iterable

import rdflib

import sbol3


class TestProperty(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_slice_assignment(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        self.assertEqual([], c.roles)
        c.roles.append(sbol3.SO_PROMOTER)
        self.assertEqual([sbol3.SO_PROMOTER], c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        c.roles.append(sbol3.SO_CDS)
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.SO_CDS], c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        c.roles[1] = sbol3.CHEBI_EFFECTOR
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.CHEBI_EFFECTOR], c.roles)
        # Make sure the underlying representation is correct
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        c.roles.append(sbol3.SO_RBS)
        expected = [sbol3.SO_PROMOTER, sbol3.CHEBI_EFFECTOR, sbol3.SO_RBS]
        self.assertEqual(expected, c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        # Replace the first two elements by slice replacement
        c.roles[0:2] = [sbol3.SO_OPERATOR, sbol3.SO_MRNA]
        expected = [sbol3.SO_OPERATOR, sbol3.SO_MRNA, sbol3.SO_RBS]
        self.assertEqual(expected, c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

    def test_boolean_property(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        c.boolean_attribute = sbol3.BooleanProperty(c, 'http://example.org#foo',
                                                    0, 1, [])
        c.boolean_attribute = True
        self.assertEqual(type(c.boolean_attribute), bool)

    def test_bounds(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        c.boolean_attribute = sbol3.BooleanProperty(c, 'http://example.org#foo',
                                                    0, math.inf, [])
        self.assertTrue(hasattr(c.boolean_attribute, '__iter__'))
        with self.assertRaises(TypeError):
            c.boolean_attribute = True
        c.int_attribute = sbol3.IntProperty(c, 'http://example.org#foo',
                                            0, math.inf, [])
        self.assertTrue(hasattr(c.int_attribute, '__iter__'))
        with self.assertRaises(TypeError):
            c.int_attribute = 0

    def test_iadd(self):
        # This is a test for the += operator, which is implemented as __iadd__()
        # When += is invoked the default __iadd__() implementation calls insert()
        # and calls set(). That confuses our auto-numbering of child object
        # identities.
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        doc = sbol3.Document()
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        doc.add(c1)
        lsc1 = sbol3.LocalSubComponent([sbol3.SBO_DNA])
        c1.features += [lsc1]
        self.assertEqual('LocalSubComponent1', lsc1.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc1.display_id),
                         lsc1.identity)
        # TODO: Add checks for document at each step below
        self.assertEqual(doc, lsc1.document)

        lsc2 = sbol3.LocalSubComponent([sbol3.SBO_DNA])
        c1.features += [lsc2]
        # Make sure lsc1 did not get renamed
        self.assertEqual('LocalSubComponent1', lsc1.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc1.display_id),
                         lsc1.identity)
        self.assertEqual(doc, lsc1.document)
        self.assertEqual('LocalSubComponent2', lsc2.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc2.display_id),
                         lsc2.identity)
        self.assertEqual(doc, lsc2.document)

        lsc3 = sbol3.LocalSubComponent([sbol3.SBO_DNA])
        lsc4 = sbol3.LocalSubComponent([sbol3.SBO_DNA])
        c1.features += [lsc3, lsc4]
        # Make sure lsc1 did not get renamed
        self.assertEqual('LocalSubComponent1', lsc1.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc1.display_id),
                         lsc1.identity)
        self.assertEqual(doc, lsc1.document)
        # Make sure lsc2 did not get renamed
        self.assertEqual('LocalSubComponent2', lsc2.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc2.display_id),
                         lsc2.identity)
        self.assertEqual(doc, lsc2.document)
        self.assertEqual('LocalSubComponent3', lsc3.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc3.display_id),
                         lsc3.identity)
        self.assertEqual(doc, lsc3.document)
        self.assertEqual('LocalSubComponent4', lsc4.display_id)
        self.assertEqual(posixpath.join(c1.identity, lsc4.display_id),
                         lsc4.identity)
        self.assertEqual(doc, lsc4.document)

    def test_not_iterable(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        self.assertIsInstance(c1.types, Iterable)
        with self.assertRaises(TypeError):
            c1.types = sbol3.SBO_PROTEIN
        with self.assertRaises(TypeError):
            c1.types = object()

    def test_attribute_name(self):
        # Verify that a property can figure out what its attribute name is.
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        types_property = c1.__dict__['types']
        aname = types_property.attribute_name
        self.assertEqual('types', aname)
        features_property = c1.__dict__['features']
        aname = features_property.attribute_name
        self.assertEqual('features', aname)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import tempfile
import unittest

import rdflib

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')
TEST_RESOURCE_DIR = os.path.join(MODULE_LOCATION, 'resources')


class TestDocument(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        logging.basicConfig(level=logging.INFO)

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_read_ntriples(self):
        # Initial test of Document.read
        filename = 'model.nt'
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 filename)
        doc = sbol3.Document()
        doc.read(test_path)
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, filename)
            doc.write(test_file, sbol3.NTRIPLES)

    def test_read_turtle(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 'model.ttl')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.TURTLE)

    def test_read_xml_model(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 'model.rdf')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.RDF_XML)
        self.assertIsNotNone(doc.find('https://sbolstandard.org/examples/toggle_switch'))
        self.assertIsNotNone(doc.find('toggle_switch'))
        self.assertIsNotNone(doc.find('https://sbolstandard.org/examples/model1'))
        # We have inferred the displayId of model1
        self.assertIsNotNone(doc.find('model1'))

    def test_read_turtle_interface(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'interface',
                                 'interface.ttl')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.TURTLE)

    def test_read_turtle_toggle_switch(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'toggle_switch',
                                 'toggle_switch.ttl')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.TURTLE)

    def test_add(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        doc = sbol3.Document()
        non_top_level = sbol3.Interface()
        with self.assertRaises(TypeError):
            doc.add(non_top_level)
        seq = sbol3.Sequence('seq1')
        doc.add(seq)
        seq2 = doc.find(seq.identity)
        self.assertEqual(seq.identity, seq2.identity)

    def test_add_multiple(self):
        # Ensure that duplicate identities cannot be added to the document.
        # See https://github.com/SynBioDex/pySBOL3/issues/39
        document = sbol3.Document()
        namespace1 = sbol3.Namespace(identity=sbol3.SBOL3_NS)
        document.add(namespace1)
        namespace2 = sbol3.Namespace(identity=sbol3.SBOL3_NS)
        with self.assertRaises(ValueError):
            document.add(namespace2)

    def test_write(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        doc = sbol3.Document()
        doc.add(sbol3.Component('c1', sbol3.SBO_DNA))
        with tempfile.TemporaryDirectory() as tmpdirname:
            nt_file = os.path.join(tmpdirname, 'test_output.nt')
            doc.write(nt_file, sbol3.NTRIPLES)
            xml_file = os.path.join(tmpdirname, 'test_output.xml')
            doc.write(xml_file, sbol3.RDF_XML)
            # No format specified for these, doc.write now figures
            # that out on its own
            out_path = os.path.join(tmpdirname, 'test_output.rdf')
            doc.write(out_path)
            out_path = os.path.join(tmpdirname, 'test_output.ttl')
            doc.write(out_path)
            out_path = os.path.join(tmpdirname, 'test_output.jsonld')
            doc.write(out_path)

    def test_bind(self):
        doc = sbol3.Document()
        prefix = 'foo'
        ns1 = 'http://example.com/foo'
        doc.bind(prefix, ns1)
        self.assertEqual(ns1, doc._namespaces[prefix])

    def test_add_namespace(self):
        doc = sbol3.Document()
        prefix = 'foo'
        ns1 = 'http://example.com/foo'
        with self.assertWarns(DeprecationWarning):
            doc.addNamespace(ns1, prefix)

    def test_guess_format(self):
        doc = sbol3.Document()
        # Expect ValueError because '.foo' file extension is unknown
        with self.assertRaises(ValueError):
            doc.read('test.foo')
        # Expect ValueError because '.foo' file extension is unknown
        with self.assertRaises(ValueError):
            doc.write('test.foo')

    def test_empty_graph(self):
        # Ensure that an empty document generates an empty RDF graph
        doc = sbol3.Document()
        g = doc.graph()
        self.assertEqual(0, len(g))

    def test_graph(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        doc = sbol3.Document()
        doc.add(sbol3.Component('foo', sbol3.SBO_DNA))
        graph = doc.graph()
        self.assertEqual(3, len(graph))
        subjects = set()
        predicates = set()
        for s, p, _ in graph:
            subjects.add(s)
            predicates.add(p)
        # Expecting 1 subject, the component
        self.assertEqual(1, len(subjects))
        # Expecting 3 predicates
        self.assertEqual(3, len(predicates))
        self.assertIn(rdflib.RDF.type, predicates)
        # Convert predicates to strings for the remaining assertions
        predicates = [str(p) for p in predicates]
        self.assertIn(sbol3.SBOL_DISPLAY_ID, predicates)
        self.assertIn(sbol3.SBOL_TYPE, predicates)

    def test_write_string(self):
        # Make sure Document.write_string produces the same output
        # as Document.write. Must use sorted NTriples for this test
        # because other formats do not have a guaranteed order.
        filename = 'model.nt'
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 filename)
        doc = sbol3.Document()
        doc.read(test_path)
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, filename)
            doc.write(test_file, sbol3.SORTED_NTRIPLES)
            with open(test_file, 'r') as infile:
                expected = infile.read()
        actual = doc.write_string(sbol3.SORTED_NTRIPLES)
        self.assertEqual(expected, actual)

    def test_validate(self):
        # Test the document level validation
        # This should validate all the objects in the document
        # and return a report containing all errors and warnings.
        doc = sbol3.Document()
        c1 = sbol3.Component('https://github.com/synbiodex/pysbol3/c1',
                             sbol3.SBO_DNA)
        doc.add(c1)
        s1 = sbol3.Sequence('https://github.com/synbiodex/pysbol3/s1')
        doc.add(s1)
        start = 10
        end = 1
        r = sbol3.Range(s1, start, end)
        sf = sbol3.SequenceFeature([r])
        c1.features.append(sf)
        report = doc.validate()
        # We should find the validation issue in the Range
        self.assertEqual(1, len(report))

    def test_find_all(self):
        test_files = {
            os.path.join(SBOL3_LOCATION, 'entity', 'model',
                         'model.nt'): 2,
            os.path.join(SBOL3_LOCATION, 'multicellular',
                         'multicellular.nt'): 75,
        }
        for file, count in test_files.items():
            doc = sbol3.Document()
            doc.read(file)
            found = doc.find_all(lambda _: True)
            self.assertEqual(count, len(found))

    def test_find_all2(self):
        file = os.path.join(SBOL3_LOCATION, 'multicellular',
                            'multicellular.nt')
        doc = sbol3.Document()
        doc.read(file)
        # Find all instances of Participation
        found = doc.find_all(lambda obj: isinstance(obj, sbol3.Participation))
        self.assertEqual(11, len(found))

    def test_visit_all(self):
        visited_list = []

        def my_visitor(obj: sbol3.Identified):
            visited_list.append(obj)
        file = os.path.join(SBOL3_LOCATION, 'multicellular',
                            'multicellular.nt')
        doc = sbol3.Document()
        doc.read(file)
        doc.accept(my_visitor)
        self.assertEqual(75, len(visited_list))

    def test_visit_participations(self):
        # Visit all and only the participations
        visited_list = []

        def my_visitor(obj: sbol3.Identified):
            if isinstance(obj, sbol3.Participation):
                visited_list.append(obj)
        file = os.path.join(SBOL3_LOCATION, 'multicellular',
                            'multicellular.nt')
        doc = sbol3.Document()
        doc.read(file)
        doc.accept(my_visitor)
        self.assertEqual(11, len(visited_list))

    def test_builder_lookup(self):
        # Test looking up a builder function
        # See https://github.com/SynBioDex/pySBOL3/issues/175
        doc = sbol3.Document()
        builder = doc.builder(sbol3.SBOL_EXTERNALLY_DEFINED)
        self.assertIsNotNone(builder)
        self.assertTrue(callable(builder))
        with self.assertRaises(ValueError):
            doc.builder('http://example.com/SomeType')
        with self.assertRaises(ValueError):
            doc.builder(None)

    def test_other_rdf_round_trip(self):
        filename = 'mixed-rdf.nt'
        test_file = os.path.join(MODULE_LOCATION, 'resources', filename)
        doc = sbol3.Document()
        doc.read(test_file)
        c_uri = 'http://example.com/sbol3/c1'
        c = doc.find(c_uri)
        self.assertIsNotNone(c)
        self.assertEqual([sbol3.SBO_DNA], c.types)
        c.types = [sbol3.SBO_PROTEIN]
        self.assertEqual([sbol3.SBO_PROTEIN], c.types)
        # Now round trip the file and make sure the
        # types property remains the same
        doc2 = sbol3.Document()
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_file = os.path.join(tmpdirname, filename)
            doc.write(tmp_file)
            doc2.read(tmp_file)
        c2 = doc2.find(c_uri)
        self.assertIsNotNone(c2)
        self.assertEqual([sbol3.SBO_PROTEIN], c2.types)

    def test_multi_type_sbol(self):
        # Load a file that includes an SBOL object that has multiple other
        # rdf:type properties
        test_file = os.path.join(TEST_RESOURCE_DIR, 'multi-type-sbol.nt')
        doc = sbol3.Document()
        doc.read(test_file)
        c_uri = 'http://example.com/sbol3/c1'
        c = doc.find(c_uri)
        self.assertIsNotNone(c)
        self.assertEqual(3, len(c._rdf_types))
        self.assertEqual(sbol3.SBOL_COMPONENT, c.type_uri)
        self.assertIsInstance(c, sbol3.Component)

    def test_multi_type_ext(self):
        # Load a file that includes an SBOL object that has multiple other
        # rdf:type properties
        test_file = os.path.join(TEST_RESOURCE_DIR, 'multi-type-ext.nt')
        doc = sbol3.Document()
        doc.read(test_file)
        uri = 'http://example.com/sbol3/c1'
        x = doc.find(uri)
        self.assertIsNotNone(x)
        self.assertEqual(3, len(x._rdf_types))
        self.assertNotEqual(sbol3.SBOL_TOP_LEVEL, x.type_uri)
        self.assertIsInstance(x, sbol3.CustomTopLevel)

    def test_multi_type_ext_builder(self):
        class MultiTypeExtension(sbol3.TopLevel):
            def __init__(self, identity: str, type_uri: str):
                super().__init__(identity=identity, type_uri=type_uri)

        def mte_builder(identity: str = None,
                        type_uri: str = None) -> sbol3.Identified:
            return MultiTypeExtension(identity=identity, type_uri=type_uri)

        ext_type_uri = 'http://example.com/fake/Type1'
        test_file = os.path.join(TEST_RESOURCE_DIR, 'multi-type-ext.nt')
        try:
            sbol3.Document.register_builder(ext_type_uri, mte_builder)
            doc = sbol3.Document()
            doc.read(test_file)
            uri = 'http://example.com/sbol3/c1'
            x = doc.find(uri)
            self.assertIsNotNone(x)
            self.assertEqual(3, len(x._rdf_types))
            self.assertEqual(ext_type_uri, x.type_uri)
            self.assertIsInstance(x, MultiTypeExtension)
        finally:
            # Reach behind the scenes to remove the registered builder
            del sbol3.Document._uri_type_map[ext_type_uri]


if __name__ == '__main__':
    unittest.main()



import abc
import math
from typing import List

from . import *


class Prefix(CustomTopLevel, abc.ABC):
    """As adopted by SBOL, om:Prefix is an abstract class that is extended
    by other classes to describe factors that are commonly represented
    by standard unit prefixes. For example, the factor 10**−3 is
    represented by the standard unit prefix "milli".

    See Appendix A Section A.2 of the SBOL 3.0 specificiation.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 factor: float, type_uri: str,
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.symbol = TextProperty(self, OM_SYMBOL, 1, 1,
                                   initial_value=symbol)
        self.alternative_symbols = TextProperty(self, OM_ALTERNATIVE_SYMBOL,
                                                0, math.inf,
                                                initial_value=alternative_symbols)
        self.label = TextProperty(self, OM_LABEL, 1, 1,
                                  initial_value=label)
        self.alternative_labels = TextProperty(self, OM_ALTERNATIVE_LABEL,
                                               0, math.inf,
                                               initial_value=alternative_labels)
        self.comment = TextProperty(self, OM_COMMENT, 0, 1,
                                    initial_value=comment)
        self.long_comment = TextProperty(self, OM_LONG_COMMENT, 0, 1,
                                         initial_value=long_comment)
        self.factor = FloatProperty(self, OM_HAS_FACTOR, 1, 1,
                                    initial_value=factor)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if not self.symbol:
            message = 'Prefix must contain a symbol'
            report.addError(self.identity, None, message)
        if not self.label:
            message = 'Prefix must contain a label'
            report.addError(self.identity, None, message)
        if not self.factor:
            message = 'Prefix must contain a factor'
            report.addError(self.identity, None, message)
        return report


class SIPrefix(Prefix):
    """The purpose of the om:SIPrefix class is to describe standard SI
    prefixes such as "milli," "centi," "kilo," etc.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 factor: float,
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_SI_PREFIX) -> None:
        super().__init__(symbol=symbol, label=label, factor=factor,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


def build_si_prefix(identity: str, *, type_uri: str = OM_SI_PREFIX) -> SBOLObject:
    obj = SIPrefix(identity=identity, type_uri=type_uri,
                   symbol=PYSBOL3_MISSING, label=PYSBOL3_MISSING, factor=1.0)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_FACTOR] = []
    return obj


Document.register_builder(OM_SI_PREFIX, build_si_prefix)


class BinaryPrefix(Prefix):
    """The purpose of the om:BinaryPrefix class is to describe standard
    binary prefixes such as "kibi," "mebi," "gibi," etc.  These
    prefixes commonly precede units of information such as "bit" and
    "byte."

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 factor: float,
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_BINARY_PREFIX) -> None:
        super().__init__(symbol=symbol, label=label, factor=factor,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


def build_binary_prefix(identity: str,
                        *, type_uri: str = OM_BINARY_PREFIX) -> SBOLObject:
    obj = BinaryPrefix(identity=identity, type_uri=type_uri,
                       symbol=PYSBOL3_MISSING, label=PYSBOL3_MISSING,
                       factor=1.0)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_FACTOR] = []
    return obj


Document.register_builder(OM_BINARY_PREFIX, build_binary_prefix)



from typing import List

import rdflib

from . import *


class CustomIdentified(Identified):

    def __init__(self, type_uri: str = None,
                 *, name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 sbol_type_uri: str = SBOL_IDENTIFIED) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self._rdf_types.append(sbol_type_uri)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if len(self._rdf_types) < 2:
            message = 'Extension classes must have at least 2 rdf:type properties'
            report.addError(self.identity, None, message)
        return report


class CustomTopLevel(TopLevel):

    def __init__(self, identity: str = None, type_uri: str = None,
                 *, attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 sbol_type_uri: str = SBOL_TOP_LEVEL) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self._rdf_types.append(sbol_type_uri)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if len(self._rdf_types) < 2:
            message = 'Extension classes must have at least 2 rdf:type properties'
            report.addError(self.identity, None, message)
        return report



import abc
import collections
from collections import MutableSequence, Iterable
from typing import Any, Optional, List, Dict, Union

from sbol3 import ValidationReport


class Property(abc.ABC):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None) -> None:
        self.property_owner = property_owner
        self.property_uri = property_uri
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        if validation_rules is None:
            validation_rules = []
        self.validation_rules = validation_rules
        # Initialize the storage for this property
        # Do not overwrite a value if already present (see issue #76)
        if self.property_uri not in self._storage():
            self._storage()[self.property_uri] = []

    def _storage(self) -> Dict[str, list]:
        return self.property_owner._properties

    @abc.abstractmethod
    def set(self, value: Any) -> None:
        pass

    @abc.abstractmethod
    def to_user(self, value: Any) -> str:
        pass

    @abc.abstractmethod
    def from_user(self, value: Any):
        pass

    def validate(self, name: str, report: ValidationReport):
        values = self._storage()[self.property_uri]
        length = len(values)
        if length < self.lower_bound:
            msg = f'Too few values for property {name}.'
            msg += f' Expected {self.lower_bound}, found {length}'
            report.addError(self.property_owner.identity, None, msg)
        elif length > self.upper_bound:
            msg = f'Too many values for property {name}.'
            msg += f' Expected {self.upper_bound}, found {length}'
            report.addError(self.property_owner.identity, None, msg)

    def item_added(self, item: Any) -> None:
        """Stub method for child classes to override if they have to do
        any additional processing on items after they are added. This method
        will be called on each individual item that was added to the list.
        """
        pass

    @property
    def attribute_name(self) -> Union[str, None]:
        """Heuristically determine which attribute is associated with
        this property. If no attribute can be found for this property
        return None.
        """
        # Iterate through the property owner's __dict__ looking for
        # an attribute whose value is this property.
        for name, value in self.property_owner.__dict__.items():
            if value is self:
                return name
        return None


class SingletonProperty(Property, abc.ABC):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None) -> None:
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound,
                         validation_rules)
        # This is a marker for SBOLObject.__get_attribute__
        self._sbol_singleton = True

    def set(self, value: Any) -> None:
        value = self.from_user(value)
        if value is None:
            if self.lower_bound == 0:
                self._storage()[self.property_uri] = []
            else:
                raise ValueError(f'Property {self.property_uri} cannot be unset')
        else:
            self._storage()[self.property_uri] = [value]
            self.item_added(value)

    def get(self) -> Any:
        try:
            value = self._storage()[self.property_uri][0]
        except IndexError:
            return None
        return self.to_user(value)


class ListProperty(Property, MutableSequence, abc.ABC):

    def __delitem__(self, key: Union[int, slice]) -> None:
        self._storage()[self.property_uri].__delitem__(key)

    def __setitem__(self, key: Union[int, slice], value: Any) -> None:
        # Do string separately because it, too, is iterable
        if isinstance(value, str):
            values = [value]
            value = self.from_user(value)
        elif isinstance(value, Iterable):
            # string is iterable so it's broken out above
            values = value
            value = [self.from_user(item) for item in value]
        else:
            # Not string or iterable
            values = [value]
            value = self.from_user(value)
        self._storage()[self.property_uri].__setitem__(key, value)
        for val in values:
            self.item_added(val)

    def __getitem__(self, key: Union[int, slice]) -> Any:
        value = self._storage()[self.property_uri].__getitem__(key)
        if isinstance(value, str):
            return self.to_user(value)
        elif isinstance(value, collections.Iterable):
            return [self.to_user(v) for v in value]
        else:
            # Not a string or an iterable, just convert
            return self.to_user(value)

    def __len__(self) -> int:
        return self._storage()[self.property_uri].__len__()

    def __contains__(self, item) -> bool:
        item = self.from_user(item)
        return self._storage()[self.property_uri].__contains__(item)

    def __eq__(self, other) -> bool:
        storage = self._storage()[self.property_uri]
        value = [self.to_user(v) for v in storage]
        return value.__eq__(other)

    def __str__(self) -> str:
        return str([self.to_user(item)
                    for item in self._storage()[self.property_uri]])

    def __repr__(self) -> str:
        return repr([self.to_user(item)
                     for item in self._storage()[self.property_uri]])

    def insert(self, index: int, value: Any) -> None:
        item = self.from_user(value)
        self._storage()[self.property_uri].insert(index, item)
        self.item_added(value)

    def set(self, value: Any) -> None:
        # The += operator (_iadd_()) does an insert followed by a set.
        # We only need one, not both. So if the current value of the
        # property is identical to the value being set, do nothing.
        if value == self._storage()[self.property_uri]:
            return
        # Give a better error message if the value is a string
        # or is not iterable.
        # See https://github.com/SynBioDex/pySBOL3/issues/195
        if (isinstance(value, (str, bytes)) or
                not isinstance(value, Iterable)):
            name = self.attribute_name
            if not name:
                name = f'Property {self.property_uri}'
            msg = f'{name} requires one or more values'
            msg += ' packed in an iterable'
            raise TypeError(msg)
        items = [self.from_user(v) for v in value]
        self._storage()[self.property_uri] = items
        for val in value:
            self.item_added(val)



import math
from typing import Union, List

from . import *


class CombinatorialDerivation(TopLevel):
    """The purpose of the CombinatorialDerivation class is to specify
    combinatorial biological designs without having to specify every
    possible design variant. For example, a CombinatorialDerivation
    can be used to specify a library of reporter gene variants that
    include different promoters and RBSs without having to specify a
    Component for every possible combination of promoter, RBS, and CDS
    in the library. Component objects that realize a
    CombinatorialDerivation can be derived in accordance with the
    class properties template, hasVariableFeature, and strategy.

    """

    def __init__(self, identity: str, template: Union[Component, str],
                 *, strategy: str = None,
                 variable_features: List[str] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_COMBINATORIAL_DERIVATION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.strategy = URIProperty(self, SBOL_STRATEGY, 0, 1,
                                    initial_value=strategy)
        self.template = ReferencedObject(self, SBOL_TEMPLATE, 1, 1,
                                         initial_value=template)
        self.variable_features = OwnedObject(self, SBOL_VARIABLE_FEATURES,
                                             0, math.inf,
                                             initial_value=variable_features,
                                             type_constraint=VariableFeature)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if self.strategy is not None:
            valid_strategies = [SBOL_ENUMERATE, SBOL_SAMPLE]
            if self.strategy not in valid_strategies:
                message = f'{self.strategy} is not a valid strategy'
                report.addError(self.identity, None, message)
        return report


def build_combinatorial_derivation(identity: str,
                                   *, type_uri: str = SBOL_COMBINATORIAL_DERIVATION):
    template = PYSBOL3_MISSING
    return CombinatorialDerivation(identity, template, type_uri=type_uri)


Document.register_builder(SBOL_COMBINATORIAL_DERIVATION, build_combinatorial_derivation)



from typing import Optional, Union, List

from . import *


class Constraint(Identified):
    """The Constraint class can be used to assert restrictions on the
    relationships of pairs of Feature objects contained by the same
    parent Component. Uses of this class include expressing
    containment (e.g., a plasmid transformed into a chassis strain),
    identity mappings (e.g., replacing a placeholder value with a
    complete definition), and expressing relative, sequence-based
    positions (e.g., the ordering of features within a template). Each
    Constraint includes the subject, object, and restriction
    properties.

    """

    def __init__(self, restriction: str, subject: Union[Identified, str],
                 object: Union[Identified, str], *, name: str = None,
                 description: str = None, derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: Optional[str] = None,
                 type_uri: str = SBOL_CONSTRAINT) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.restriction = URIProperty(self, SBOL_RESTRICTION, 1, 1,
                                       initial_value=restriction)
        self.subject = ReferencedObject(self, SBOL_SUBJECT, 1, 1,
                                        initial_value=subject)
        self.object = ReferencedObject(self, SBOL_OBJECT, 1, 1,
                                       initial_value=object)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if not self.restriction:
            message = 'Constraint must have a restriction'
            report.addError(self.identity, None, message)
        if not self.subject:
            message = 'Constraint must have a subject'
            report.addError(self.identity, None, message)
        if not self.object:
            message = 'Constraint must have an object'
            report.addError(self.identity, None, message)
        return report


def build_constraint(identity: str, type_uri: str = SBOL_CONSTRAINT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Constraint(missing, missing, missing, identity=identity, type_uri=type_uri)
    obj._properties[SBOL_RESTRICTION] = []
    obj._properties[SBOL_SUBJECT] = []
    obj._properties[SBOL_OBJECT] = []
    return obj


Document.register_builder(SBOL_CONSTRAINT, build_constraint)



import os
import posixpath
import unittest
from collections import Container

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestComponent(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_roles(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        self.assertListEqual([], list(c.roles))
        self.assertEqual([], c.roles)
        c.roles = [sbol3.SO_PROMOTER]
        # Make sure the underlying datatype did not get overwritten
        self.assertNotIsInstance(c.roles, list)
        self.assertIsInstance(c.roles, Container)
        self.assertIn(sbol3.SO_PROMOTER, c.roles)
        self.assertEqual([sbol3.SO_PROMOTER], c.roles)
        self.assertTrue(c.roles == [sbol3.SO_PROMOTER])
        self.assertTrue([sbol3.SO_PROMOTER] == c.roles)
        # Other list manipulations
        c.roles.append(sbol3.SO_CDS)
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.SO_CDS], c.roles)
        self.assertEqual([sbol3.SO_CDS], c.roles[1:])
        c.roles[1:] = [sbol3.SO_RBS]
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.SO_RBS], c.roles)
        self.assertEqual([sbol3.SO_RBS], c.roles[1:])

    def test_features(self):
        # See https://github.com/SynBioDex/pySBOL3/issues/149
        # Note: this example was modified when fixing
        #       https://github.com/SynBioDex/pySBOL3/issues/178
        #       media_variable is unused in the original example so
        #       it has been commented out here

        media_template = sbol3.LocalSubComponent(types=[sbol3.SBO_FUNCTIONAL_ENTITY])
        media_template.name = 'media template'

        # variable_uri = 'https://github.com/synbiodex/pysbol3/variable'
        # media_variable = sbol3.VariableFeature(cardinality=sbol3.SBOL_ONE,
        #                                        variable=media_template)
        # media_variable.variable = media_template

        all_sample_templates = [media_template]
        sample_template_uri = 'https://sd2e.org/measurement_template'
        sample_template = sbol3.Component(identity=sample_template_uri,
                                          types=sbol3.SBO_FUNCTIONAL_ENTITY)
        sample_template.name = 'measurement template'
        sample_template.features = all_sample_templates
        self.assertEqual(1, len(sample_template.features))
        self.assertEqual(media_template.identity,
                         sample_template.features[0].identity)

    def test_type_validation(self):
        # Test the validation of types on owned object properties by
        # going behind the scenes to set a bad value
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        report = c1.validate()
        self.assertEqual(0, len(report))
        c1._owned_objects[sbol3.SBOL_FEATURES] = [sbol3.Interface(),
                                                  sbol3.Implementation('i1')
                                                  ]
        report = c1.validate()
        # Expecting 2 errors, one for each inappropriate value
        self.assertEqual(2, len(report))

    def test_cloning(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        new_identity = 'c2'
        c2 = c1.clone(new_identity)
        self.assertEqual(posixpath.join(sbol3.get_namespace(), new_identity),
                         c2.identity)

    def test_cloning_with_references(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        new_identity = 'c2'
        seq1 = sbol3.Sequence('s1')
        c1.sequences.append(seq1)
        c2 = c1.clone(new_identity)
        self.assertEqual(posixpath.join(sbol3.get_namespace(), new_identity),
                         c2.identity)
        self.assertListEqual(list(c1.sequences), list(c2.sequences))

    def test_cloning_with_children(self):
        # This test does not use `sbol3.set_namespace` as the other
        # cloning unit tests do. This is on purpose to verify that
        # cloning does not rely on the default namespace.
        doc = sbol3.Document()
        namespace = 'https://github.com/synbiodex/pysbol3'
        c1_identity = posixpath.join(namespace, 'c1')
        c2_identity = posixpath.join(namespace, 'c2')
        s1_identity = posixpath.join(namespace, 's1')
        c1 = sbol3.Component(c1_identity, sbol3.SBO_DNA)
        doc.add(c1)
        seq1 = sbol3.Sequence(s1_identity)
        doc.add(seq1)
        c1.sequences.append(seq1)
        sc1 = sbol3.SubComponent(c1)
        es1 = sbol3.EntireSequence(seq1)
        sc1.source_locations.append(es1)
        c1.features = [sc1]
        c2 = c1.clone(c2_identity)
        self.assertEqual(c2_identity, c2.identity)
        self.assertIsNone(c2.document)
        # Check on the SubComponent
        sc2 = c2.features[0]
        self.assertIsInstance(sc2, sbol3.SubComponent)
        self.assertNotEqual(sc1.identity, sc2.identity)
        self.assertTrue(sc2.identity.startswith(c2.identity))
        # Ensure that the reference was updated properly
        self.assertEqual(c2.identity, sc2.instance_of)
        self.assertIsNone(sc2.document)
        es2 = sc2.source_locations[0]
        self.assertIsInstance(es2, sbol3.EntireSequence)
        self.assertNotEqual(es1.identity, es2.identity)
        self.assertTrue(es2.identity.startswith(c2.identity))
        self.assertEqual(es1.sequence, es2.sequence)
        self.assertIsNone(es2.document)

    def test_cloning_references(self):
        # Verify that when we clone the constraint references
        # are in the namespace of the clone
        test_file = os.path.join(SBOL3_LOCATION, 'toggle_switch',
                                 'toggle_switch.nt')
        doc = sbol3.Document()
        doc.read(test_file)
        toggle_uri = 'https://sbolstandard.org/examples/toggle_switch'
        toggle = doc.find(toggle_uri)
        new_uri = 'https://github.com/synbiodex/pysbol3/toggle_switch'
        toggle_clone = toggle.clone(new_uri)
        self.assertNotEqual(toggle.identity, toggle_clone.identity)
        self.assertEqual(new_uri, toggle_clone.identity)
        doc2 = sbol3.Document()
        doc2.add(toggle_clone)
        self.assertEqual(len(toggle.constraints), len(toggle_clone.constraints))
        for i in range(len(toggle.constraints)):
            c = toggle.constraints[i]
            c_clone = toggle_clone.constraints[i]
            self.assertNotEqual(c.identity, c_clone.identity)
            s = c.subject.lookup()
            self.assertIsInstance(s, sbol3.ComponentReference)
            self.assertTrue(s.identity.startswith(toggle.identity))
            s_clone = c_clone.subject.lookup()
            self.assertIsInstance(s_clone, sbol3.ComponentReference)
            self.assertTrue(s_clone.identity.startswith(toggle_clone.identity))
            self.assertNotEqual(s.identity, s_clone.identity)
            self.assertEqual(s.feature, s_clone.feature)
            o = c.object.lookup()
            self.assertIsInstance(o, sbol3.ComponentReference)
            self.assertTrue(o.identity.startswith(toggle.identity))
            o_clone = c_clone.object.lookup()
            self.assertIsInstance(o_clone, sbol3.ComponentReference)
            self.assertTrue(o_clone.identity.startswith(toggle_clone.identity))
            self.assertNotEqual(o.identity, o_clone.identity)
            self.assertEqual(o.feature, o_clone.feature)


if __name__ == '__main__':
    unittest.main()



from typing import List

from . import *


class Sequence(TopLevel):

    def __init__(self, identity: str, *,
                 elements: str = None, encoding: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None, generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_SEQUENCE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.elements = TextProperty(self, SBOL_ELEMENTS, 0, 1,
                                     initial_value=elements)
        self.encoding = URIProperty(self, SBOL_ENCODING, 0, 1,
                                    initial_value=encoding)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        # If sequence is set, encoding is REQUIRED
        if self.elements and not self.encoding:
            message = 'Sequence encoding is required if elements are set'
            report.addError(self.identity, None, message)
        # Check that the encoding is part of the recommended set
        encodings = {
            IUPAC_DNA_ENCODING, IUPAC_RNA_ENCODING, IUPAC_PROTEIN_ENCODING,
            INCHI_ENCODING, SMILES_ENCODING
        }
        if self.encoding and self.encoding not in encodings:
            message = 'Sequence encoding is not in the recommended set'
            report.addWarning(self.identity, 'sbol3-10505', message)
        return report


Document.register_builder(SBOL_SEQUENCE, Sequence)



import math
from typing import List

from . import *


class Interface(Identified):
    """The Interface class is a way of explicitly specifying the interface
    of a Component.

    """

    def __init__(self, *, input: str = None, output: str = None,
                 nondirectional: str = None, name: str = None,
                 description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_INTERFACE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.input = ReferencedObject(self, SBOL_INPUT, 0, math.inf,
                                      initial_value=input)
        self.output = ReferencedObject(self, SBOL_OUTPUT, 0, math.inf,
                                       initial_value=output)
        self.nondirectional = ReferencedObject(self, SBOL_NONDIRECTIONAL,
                                               0, math.inf,
                                               initial_value=nondirectional)


def build_interface(identity: str, *, type_uri: str = SBOL_INTERFACE) -> SBOLObject:
    return Interface(identity=identity, type_uri=type_uri)


Document.register_builder(SBOL_INTERFACE, build_interface)



import collections
import logging
import warnings
from typing import Dict, Callable, List, Optional

import rdflib
# Get the rdflib-jsonld capability initialized
# Note: this is for side effect. The parser is not used.
# The side effect is that the JSON-LD parser is registered in RDFlib.
from rdflib_jsonld import parser as jsonld_parser

from . import *
from .object import BUILDER_REGISTER

_default_bindings = {
    'sbol': SBOL3_NS,
    'prov': PROV_NS,
    'om': OM_NS,
    # Should others like SO, SBO, and CHEBI be added?
}


class Document:

    @staticmethod
    def register_builder(type_uri: str,
                         builder: Callable[[str, str], SBOLObject]) -> None:
        """A builder function will be called with an identity and a
        keyword argument type_uri.

        builder(identity_uri: str, type_uri: str = None) -> SBOLObject
        """
        Document._uri_type_map[type_uri] = builder

    # Map type URIs to a builder function to construct entities from
    # RDF triples.
    _uri_type_map: Dict[str, Callable[[str, str], Identified]] = BUILDER_REGISTER

    def __init__(self):
        self.logger = logging.getLogger(SBOL_LOGGER_NAME)
        self.objects: List[Identified] = []
        # Orphans are non-TopLevel objects that are not otherwise linked
        # into the object hierarchy.
        self.orphans: List[Identified] = []
        self._namespaces: Dict[str, str] = _default_bindings.copy()
        # Non-SBOL triples. These are triples that are not recognized as
        # SBOL. They are stored in _other_rdf for round-tripping purposes.
        self._other_rdf = rdflib.Graph()

    def _build_extension_object(self, identity: str, sbol_type: str,
                                types: List[str]) -> Optional[Identified]:
        custom_types = {
            SBOL_IDENTIFIED: CustomIdentified,
            SBOL_TOP_LEVEL: CustomTopLevel
        }
        if sbol_type not in custom_types:
            msg = f'{identity} has SBOL type {sbol_type} which is not one of'
            msg += f' {custom_types.keys()}. (See Section 6.11)'
            raise SBOLError(msg)
        # Look for a builder associated with one of the rdf:types.
        # If none of the rdf:types have a builder, use the sbol_type's builder
        builder = None
        build_type = None
        for type_uri in types:
            try:
                builder = self._uri_type_map[type_uri]
                build_type = type_uri
                break
            except KeyError:
                logging.warning(f'No builder found for {type_uri}')
        if builder is None:
            builder = custom_types[sbol_type]
            build_type = types[0]
        return builder(identity=identity, type_uri=build_type)

    def _build_object(self, identity: str, types: List[str]) -> Optional[Identified]:
        # Given an identity and a list of RDF types, build an object if possible.
        # If there is 1 SBOL type and we don't know it, raise an exception
        # If there are multiple types and 1 is TopLevel or Identified, then
        #    it is an extension. Use the other types to try to build it. If
        #    no other type is known, build a generic TopLevel or Identified.
        sbol_types = [t for t in types if t.startswith(SBOL3_NS)]
        if len(sbol_types) == 0:
            # If there are no SBOL types in the list. Ignore this entity.
            # Its triples will be stored in self._other_rdf later in the
            # load process.
            return None
        if len(sbol_types) > 1:
            # If there are multiple SBOL types in the list, raise an error.
            # SBOL 3.0.1 Section 5.4: "an object MUST have no more than one
            # rdfType property in the 'http://sbols.org/v3#' namespace"
            msg = f'{identity} has more than one rdfType property in the'
            msg += f' {SBOL3_NS} namespace.'
            raise SBOLError(msg)
        extension_types = {
            SBOL_IDENTIFIED: CustomIdentified,
            SBOL_TOP_LEVEL: CustomTopLevel
        }
        sbol_type = sbol_types[0]
        if sbol_type in extension_types:
            # Build an extension object
            types.remove(sbol_type)
            return self._build_extension_object(identity, sbol_type, types)
        else:
            try:
                builder = self._uri_type_map[sbol_type]
            except KeyError:
                logging.warning(f'No builder found for {sbol_type}')
                raise SBOLError(f'Unknown type {sbol_type}')
            obj = builder(identity=identity, type_uri=sbol_type)
            return obj

    def _parse_objects(self, graph: rdflib.Graph) -> Dict[str, SBOLObject]:
        # First extract the identities and their types. Each identity
        # can have either one or two rdf:type properties. If one,
        # create the entity. If two, it is a custom type (see section
        # 6.11 of the spec) and we instantiate it specially.
        #
        identity_types: Dict[str, List[str]] = collections.defaultdict(list)
        for s, p, o in graph.triples((None, rdflib.RDF.type, None)):
            str_o = str(o)
            str_s = str(s)
            identity_types[str_s].append(str_o)
        # Now iterate over the identity->type dict creating the objects.
        result = {}
        for identity, types in identity_types.items():
            obj = self._build_object(identity, types)
            if obj:
                obj.document = self
                result[obj.identity] = obj
        return result

    @staticmethod
    def _parse_attributes(objects, graph):
        for s, p, o in graph.triples((None, None, None)):
            str_s = str(s)
            str_p = str(p)
            try:
                obj = objects[str_s]
            except KeyError:
                # Object is not an SBOL object, skip it
                continue
            if str_p in obj._owned_objects:
                other_identity = str(o)
                other = objects[other_identity]
                obj._owned_objects[str_p].append(other)
            elif str_p == RDF_TYPE:
                # Handle rdf:type specially because the main type(s)
                # will already be in the list from the build_object
                # phase and those entries need to be maintained and
                # we don't want duplicates
                if o not in obj._properties[str_p]:
                    obj._properties[str_p].append(o)
            else:
                obj._properties[str_p].append(o)

    @staticmethod
    def _clean_up_singletons(objects: Dict[str, SBOLObject]):
        """Clean up singleton properties after reading an SBOL file.

        When an SBOL file is read, values are appended to the property
        stores without knowledge of which stores are singletons and
        which stores are lists. This method cleans up singleton properties
        by ensuring that each has exactly one value.
        """
        # This is necessary due to defaulting of properties when using
        # the builder. Some objects have required properties, which the
        # builder sets. In the case of singleton values, that can result
        # in multiple values in a singleton property. Only the first value
        # is used, so the value read from file is ignored.
        for _, obj in objects.items():
            for name, attr in obj.__dict__.items():
                if isinstance(attr, SingletonProperty):
                    prop_uri = attr.property_uri
                    store = attr._storage()
                    if len(store[prop_uri]) > 1:
                        store[prop_uri] = store[prop_uri][-1:]

    def clear(self) -> None:
        self.objects.clear()
        self._namespaces = _default_bindings.copy()

    def _parse_graph(self, graph) -> None:
        objects = self._parse_objects(graph)
        self._parse_attributes(objects, graph)
        self._clean_up_singletons(objects)
        # Validate all the objects
        # TODO: Where does this belong? Is this automatic?
        #       Or should a user invoke validate?
        # for obj in objects.values():
        #     obj.validate()
        # Store the TopLevel objects in the Document
        self.objects = [obj for uri, obj in objects.items()
                        if isinstance(obj, TopLevel)]
        # Gather Orphans for future writing.
        # These are expected to be non-TopLevel annotation objects whose owners
        # have no custom implementation (i.e. no builder registered). These objects
        # will be written out as part of Document.write_string()
        self.orphans = []
        for uri, obj in objects.items():
            if obj in self.objects:
                continue
            found = self.find(uri)
            if found:
                continue
            self.orphans.append(obj)
        # Store the namespaces in the Document for later use
        for prefix, uri in graph.namespaces():
            self.bind(prefix, uri)
        # Remove the triples for every object we have loaded, leaving
        # the non-RDF triples for round tripping
        # See https://github.com/SynBioDex/pySBOL3/issues/96
        for uri in objects:
            graph.remove((rdflib.URIRef(uri), None, None))
        # Now tuck away the graph for use in Document.write_string()
        self._other_rdf = graph

    def _guess_format(self, fpath: str):
        return rdflib.util.guess_format(fpath)

    # Formats: 'n3', 'nt', 'turtle', 'xml'
    def read(self, location: str, file_format: str = None) -> None:
        if file_format is None:
            file_format = self._guess_format(location)
        if file_format is None:
            raise ValueError('Unable to determine file format')
        if file_format == SORTED_NTRIPLES:
            file_format = NTRIPLES
        graph = rdflib.Graph()
        graph.load(location, format=file_format)
        return self._parse_graph(graph)

    # Formats: 'n3', 'nt', 'turtle', 'xml'
    def read_string(self, data: str, file_format: str) -> None:
        # TODO: clear the document, this isn't append
        if file_format == SORTED_NTRIPLES:
            file_format = NTRIPLES
        graph = rdflib.Graph()
        graph.parse(data=data, format=file_format)
        return self._parse_graph(graph)

    def add(self, obj: TopLevel) -> None:
        """Add objects to the document.
        """
        if not isinstance(obj, TopLevel):
            message = f'Expected TopLevel instance, {type(obj).__name__} found'
            raise TypeError(message)
        found_obj = self.find(obj.identity)
        if found_obj is not None:
            message = f'An entity with identity "{obj.identity}"'
            message += ' already exists in document'
            raise ValueError(message)
        self.objects.append(obj)

        # Assign this document to the object tree rooted
        # in the TopLevel being added
        def assign_document(x: Identified):
            x.document = self
        obj.accept(assign_document)

    def _find_in_objects(self, search_string: str) -> Optional[Identified]:
        # TODO: implement recursive search
        for obj in self.objects:
            # TODO: needs an object.find(search_string) method on ... Identified?
            result = obj.find(search_string)
            if result is not None:
                return result
        return None

    def find(self, search_string: str) -> Optional[Identified]:
        """Find an object by identity URI or by display_id.

        :param search_string: Either an identity URI or a display_id
        :type search_string: str
        :returns: The named object or ``None`` if no object was found

        """
        for obj in self.objects:
            if obj.identity == search_string:
                return obj
            if obj.display_id and obj.display_id == search_string:
                return obj
        return self._find_in_objects(search_string)

    def write_string(self, file_format: str) -> str:
        graph = self.graph()
        if file_format == SORTED_NTRIPLES:
            # have RDFlib give us the ntriples as a string
            nt_text = graph.serialize(format='nt')
            # split it into lines
            lines = nt_text.splitlines(keepends=True)
            # sort those lines
            lines.sort()
            # write out the lines
            # RDFlib gives us bytes, so open file in binary mode
            result = b''.join(lines)
        elif file_format == JSONLD:
            context = {f'@{prefix}': uri for prefix, uri in self._namespaces.items()}
            context['@vocab'] = 'https://sbolstandard.org/examples/'
            result = graph.serialize(format=file_format, context=context)
        else:
            result = graph.serialize(format=file_format)
        return result.decode()

    def write(self, fpath: str, file_format: str = None) -> None:
        """Write the document to file.

        If file_format is None the desired format is guessed from the
        extension of fpath. If file_format cannot be guessed a ValueError
        is raised.
        """
        if file_format is None:
            file_format = self._guess_format(fpath)
        if file_format is None:
            raise ValueError('Unable to determine file format')
        with open(fpath, 'w') as outfile:
            outfile.write(self.write_string(file_format))

    def graph(self) -> rdflib.Graph:
        """Convert document to an RDF Graph.

        The returned graph is a snapshot of the document and will
        not be updated by subsequent changes to the document.
        """
        graph = rdflib.Graph()
        for prefix, uri in self._namespaces.items():
            graph.bind(prefix, uri)
        for orphan in self.orphans:
            orphan.serialize(graph)
        for obj in self.objects:
            obj.serialize(graph)
        # Add the non-SBOL RDF triples into the generated graph
        graph += self._other_rdf
        return graph

    def bind(self, prefix: str, uri: str) -> None:
        """Bind a prefix to an RDF namespace in the written RDF document.

        These prefixes make the written RDF easier for humans to read.
        These prefixes do not change the semantic meaning of the RDF
        document in any way.
        """
        # Remove any prefix referencing the given URI
        if uri in self._namespaces.values():
            for k, v in list(self._namespaces.items()):
                if v == uri:
                    del self._namespaces[k]
        self._namespaces[prefix] = uri

    def addNamespace(self, namespace: str, prefix: str) -> None:
        """Document.addNamespace is deprecated. Replace with Document.bind.

        Document.addNamespace existed in pySBOL2 and was commonly used.

        Document.addNamespace(namespace, prefix) should now be
        Document.bind(prefix, namespace). Note the change of argument
        order.
        """
        warnings.warn('Use Document.bind() instead', DeprecationWarning)
        self.bind(prefix, namespace)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        """Validate all objects in this document."""
        if report is None:
            report = ValidationReport()
        for obj in self.objects:
            obj.validate(report)
        return report

    def find_all(self, predicate: Callable[[Identified], bool]) -> List[Identified]:
        """Executes a predicate on every object in the document tree,
        gathering the list of objects to which the predicate returns true.
        """
        result: List[Identified] = []

        def wrapped_filter(visited: Identified):
            if predicate(visited):
                result.append(visited)
        self.accept(wrapped_filter)
        return result

    # TODO: Implement the visitor pattern allowing the user to specify
    #       a function to direct the search, instead of only doing
    #       depth-first search. For example, visit objects in provenance
    #       order using Identified.derived_from.
    # def accept_with_strategy(self, visitor: Callable, *, strategy: Callable = None):
    #     # Use the strategy to compute the set of objects to visit
    #     if strategy is None:
    #         visit_list = self.objects
    #     else:
    #         visit_list = self.find_all(strategy)
    #     while visit_list:
    #         for obj in visit_list:
    #             obj.accept(visitor)
    #         # Compute the next set of objects to visit
    #         if strategy is None:
    #             visit_list = []
    #         else:
    #             visit_list = self.find_all(strategy)

    def accept(self, visitor: Callable[[Identified], None]):
        """Implement the visitor pattern by invoking `visitor` on all
        top-level objects in the document. Those objects, in turn, will
        invoke `visitor` on all of their child objects.
        """
        for obj in self.objects:
            obj.accept(visitor)

    def builder(self, type_uri: str) -> Callable[[str, str], Identified]:
        """Lookup up the builder callable for the given type_uri.

        The builder must have been previously registered under this
        type_uri via Document.register_builder().

        :raises: ValueError if the type_uri does not have an associated
                 builder.
        """
        try:
            return self._uri_type_map[type_uri]
        except KeyError:
            raise ValueError(f'No builder for {type_uri}')



import unittest

import sbol3


class TestVariableComponent(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        vc = sbol3.VariableFeature(cardinality=sbol3.SBOL_ZERO_OR_MORE,
                                   variable=sbol3.PYSBOL3_MISSING)
        self.assertIsNotNone(vc)
        # Verify the correct default values
        self.assertEqual(sbol3.SBOL_ZERO_OR_MORE, vc.cardinality)
        self.assertEqual(sbol3.PYSBOL3_MISSING, vc.variable)
        self.assertEqual(sbol3.SBOL_VARIABLE_FEATURE, vc.type_uri)

    def test_invalid_create(self):
        my_cardinality = 'https://github.com/synbiodex/pysbol3#someNumber'
        vf = sbol3.VariableFeature(cardinality=my_cardinality,
                                   variable=sbol3.PYSBOL3_MISSING)
        report = vf.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))

    def test_round_trip1(self):
        # See https://github.com/SynBioDex/pySBOL3/issues/155
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        doc1 = sbol3.Document()
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        doc1.add(comp1)
        cd1 = sbol3.CombinatorialDerivation('cd1', comp1)
        self.assertEqual(comp1.identity, cd1.template)
        doc1.add(cd1)
        vf1 = sbol3.VariableFeature(cardinality=sbol3.SBOL_ONE,
                                    variable=sbol3.PYSBOL3_MISSING)
        cd1.variable_features.append(vf1)
        self.assertTrue(vf1.identity.startswith(cd1.identity))
        doc2 = sbol3.Document()
        doc2.read_string(doc1.write_string(sbol3.SORTED_NTRIPLES),
                         sbol3.SORTED_NTRIPLES)
        comp2 = doc2.find(comp1.identity)
        self.assertIsInstance(comp2, sbol3.Component)
        cd2 = doc2.find(cd1.identity)
        self.assertIsInstance(cd2, sbol3.CombinatorialDerivation)
        vf2 = doc2.find(vf1.identity)
        self.assertIsInstance(vf2, sbol3.VariableFeature)

    def test_round_trip2(self):
        # See https://github.com/SynBioDex/pySBOL3/issues/159
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        doc1 = sbol3.Document()
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        doc1.add(comp1)
        cd1 = sbol3.CombinatorialDerivation('cd1', comp1)
        self.assertEqual(comp1.identity, cd1.template)
        doc1.add(cd1)
        vf1 = sbol3.VariableFeature(cardinality=sbol3.SBOL_ONE,
                                    variable=sbol3.PYSBOL3_MISSING)
        hour = 'https://identifiers.org/ncit:C25529'
        m1 = sbol3.Measure(32, hour)
        vf1.variant_measures.append(m1)
        cd1.variable_features.append(vf1)
        self.assertTrue(vf1.identity.startswith(cd1.identity))
        # Ensure that Measure m1 is valid. The bug tested here was that it
        # had been assigned an invalid displayId.
        report = m1.validate()
        self.assertEqual(0, len(report.errors))


if __name__ == '__main__':
    unittest.main()



import math
from typing import List, Union

from . import *


class Component(TopLevel):

    def __init__(self, identity: str, types: Union[List[str], str],
                 *, roles: List[str] = None,
                 sequences: List[str] = None,
                 features: List[Feature] = None,
                 constraints: List[Constraint] = None,
                 interactions: List[Interaction] = None,
                 interfaces: List[Interface] = None,
                 models: List[str] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None, generated_by: List[str] = None,
                 measures: List[SBOLObject] = None, type_uri: str = SBOL_COMPONENT):
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        if isinstance(types, str):
            types = [types]
        self.types: Union[List, Property] = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                                        initial_value=types)
        self.roles = URIProperty(self, SBOL_ROLE, 0, math.inf,
                                 initial_value=roles)
        self.sequences = ReferencedObject(self, SBOL_SEQUENCES, 0, math.inf,
                                          initial_value=sequences)
        self.features = OwnedObject(self, SBOL_FEATURES, 0, math.inf,
                                    initial_value=features,
                                    type_constraint=Feature)
        self.interactions = OwnedObject(self, SBOL_INTERACTIONS, 0, math.inf,
                                        initial_value=interactions,
                                        type_constraint=Interaction)
        self.constraints = OwnedObject(self, SBOL_CONSTRAINTS, 0, math.inf,
                                       initial_value=constraints,
                                       type_constraint=Constraint)
        self.interfaces = OwnedObject(self, SBOL_INTERFACES, 0, 1,
                                      initial_value=interfaces,
                                      type_constraint=Interface)
        self.models = ReferencedObject(self, SBOL_MODELS, 0, math.inf,
                                       initial_value=models)


def build_component(identity: str, *, type_uri: str = SBOL_COMPONENT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Component(identity, [missing], type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_TYPE] = []
    return obj


Document.register_builder(SBOL_COMPONENT, build_component)



import math

from . import *
# Feature is not exported
from .feature import Feature
from .typing import *


class LocalSubComponent(Feature):
    """LocalSubComponent serves as a way to create a placeholder in more
    complex Components, such as a variable to be filled in later or a
    composite that exists only within the context of the parent
    Component.

    """

    def __init__(self, types: List[str],
                 *, locations: List[Location] = None,
                 roles: List[str] = None, orientation: str = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_LOCAL_SUBCOMPONENT) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         roles=roles, orientation=orientation, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.types: uri_list = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                           initial_value=types)
        self.locations = OwnedObject(self, SBOL_LOCATION, 0, math.inf,
                                     initial_value=locations,
                                     type_constraint=Location)


def build_local_subcomponent(identity: str,
                             *, type_uri: str = SBOL_LOCAL_SUBCOMPONENT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = LocalSubComponent([missing], identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_TYPE] = []
    return obj


Document.register_builder(SBOL_LOCAL_SUBCOMPONENT, build_local_subcomponent)



import math

from . import *
from .typing import *


class Participation(Identified):
    """Each Participation represents how a particular Feature behaves in
    its parent Interaction.

    """

    def __init__(self, roles: List[str],
                 participant: Union[SBOLObject, str],
                 *, name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_PARTCIPATION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.roles: uri_list = URIProperty(self, SBOL_ROLE, 1, math.inf,
                                           initial_value=roles)
        self.participant = ReferencedObject(self, SBOL_PARTICIPANT, 1, 1,
                                            initial_value=participant)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if len(self.roles) < 1:
            message = 'Participation must have at least one role'
            report.addError(self.identity, None, message)
        if self.participant is None:
            message = 'Participation must have a participant'
            report.addError(self.identity, None, message)
        return report


def build_participation(identity: str,
                        *, type_uri: str = SBOL_PARTCIPATION) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Participation(roles=[missing], participant=missing,
                        identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_ROLE] = []
    obj._properties[SBOL_PARTICIPANT] = []
    return obj


Document.register_builder(SBOL_PARTCIPATION, build_participation)



import copy
import math
from typing import List, Dict, Callable

from . import *


class TopLevel(Identified):
    """TopLevel is an abstract class that is extended by any Identified
    class that can be found at the top level of an SBOL document or
    file. In other words, TopLevel objects are not nested inside any
    other object via composite aggregation. Instead of nesting,
    composite TopLevel objects refer to subordinate TopLevel objects
    by their URIs using shared aggregation.

    """

    def __init__(self, identity: str, type_uri: str,
                 *, attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None) -> None:
        # Sanity check identity, which is required for a TopLevel
        # More checking on identity happens in Identified, but Identified
        # does not require an identity, only TopLevel does.
        if not identity or not isinstance(identity, str):
            raise ValueError('Identity must be a non-empty string')
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.attachments = ReferencedObject(self, SBOL_HAS_ATTACHMENT, 0, math.inf,
                                            initial_value=attachments)

    def validate_identity(self, report: ValidationReport) -> None:
        # TODO: See section 5.1 for rules about identity for TopLevel
        pass

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        return report

    def _reparent_child_objects(self) -> None:
        for k, v in self.__dict__.items():
            if not isinstance(v, Property):
                continue
            if v.property_uri not in self._owned_objects:
                continue
            old_list = self._owned_objects[v.property_uri]
            if len(old_list) == 0:
                continue
            self._owned_objects[v.property_uri] = []
            if isinstance(v, ListProperty):
                setattr(self, k, old_list)
            else:
                setattr(self, k, old_list[0])

    def clone(self, new_identity: str) -> 'TopLevel':
        obj = copy.deepcopy(self)
        identity_map = {self.identity: obj}
        # Set identity of new object
        obj._identity = obj._make_identity(new_identity)
        # Set display_id of new object
        obj._display_id = obj._extract_display_id(obj._identity)
        # Drop the document pointer
        obj.document = None

        # Erase the identity and display_id of the entire tree
        obj.accept(make_erase_identity_visitor(identity_map))

        # Now remove and re-add all child objects to update
        # their identities
        obj._reparent_child_objects()

        # TODO: Now remap any properties that reference old
        #       identities in the identity_map
        obj.accept(make_update_references_visitor(identity_map))
        return obj


def make_erase_identity_visitor(identity_map: Dict[str, Identified])\
        -> Callable[[Identified], None]:
    def erase_identity_visitor(x):
        if isinstance(x, TopLevel):
            return
        identity_map[x.identity] = x
        x._identity = None
        x._display_id = None
        x.document = None
    return erase_identity_visitor


def make_update_references_visitor(identity_map: Dict[str, Identified])\
        -> Callable[[Identified], None]:
    def update_references_visitor(x):
        # Use the identity map to update references.
        # References to objects outside of the object
        # being cloned will be left as is.
        for k, v in x.__dict__.items():
            if not isinstance(v, Property):
                continue
            if v.property_uri not in x._properties:
                continue
            items = x._properties[v.property_uri]
            for i in range(len(items)):
                str_item = str(items[i])
                if str_item in identity_map:
                    new_reference = identity_map[str_item].identity
                    # The item is probably an rdflib.URIRef. We take
                    # the type of the item and use that type to
                    # construct a new instance of the same type.
                    # Hacky? yes, and it seems to work
                    constructor = type(items[i])
                    items[i] = constructor(new_reference)
    return update_references_visitor



import math

from . import *

# Feature is not exported
from .feature import Feature
from .typing import *


class ExternallyDefined(Feature):
    """The ExternallyDefined class has been introduced so that
    external definitions in databases like ChEBI or UniProt
    can be referenced.

    """

    def __init__(self, types: List[str], definition: str,
                 *, roles: List[str] = None, orientation: str = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = SBOL_EXTERNALLY_DEFINED):
        super().__init__(identity=identity, type_uri=type_uri,
                         roles=roles, orientation=orientation, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.types: uri_list = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                           initial_value=types)
        self.definition: uri_singleton = URIProperty(self, SBOL_DEFINITION, 1, 1,
                                                     initial_value=definition)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if len(self.types) < 1:
            message = 'ExternallyDefined must contain at least 1 type'
            report.addError(self.identity, None, message)
        if self.definition is None:
            message = 'ExternallyDefined must have a definition'
            report.addError(self.identity, None, message)
        return report


def build_externally_defined(identity: str,
                             *, type_uri: str = SBOL_EXTERNALLY_DEFINED) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = ExternallyDefined([missing], missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_TYPE] = []
    obj._properties[SBOL_DEFINITION] = []
    return obj


Document.register_builder(SBOL_EXTERNALLY_DEFINED, build_externally_defined)



import unittest

from typing import Union

import sbol3

SAMPLE_SET_URI = 'http://bioprotocols.org/opil/v1#SampleSet'


class SampleSet(sbol3.CombinatorialDerivation, sbol3.CustomTopLevel):

    def __init__(self, identity: str, template: Union[sbol3.Component, str],
                 *, type_uri: str = SAMPLE_SET_URI):
        super().__init__(identity=identity, template=template,
                         type_uri=type_uri)
        # Add additional properties here


def build_sample_set(identity: str,
                     *, type_uri: str = SAMPLE_SET_URI):
    template = sbol3.PYSBOL3_MISSING
    return SampleSet(identity=identity, template=template, type_uri=type_uri)


sbol3.Document.register_builder(SAMPLE_SET_URI, build_sample_set)


class TestExtension(unittest.TestCase):

    def test_sample_set(self):
        # Test extending an SBOL class to create a new type with all the
        # properties of the SBOL type plus whatever other properties the
        # programmer defines. The key to doing this properly is to use
        # multiple inheritance to extend both the desired SBOL type and
        # CustomTopLevel so that the type is properly serialized to RDF.
        doc = sbol3.Document()
        ss_uri = 'http://example.org/sbol3/ss1'
        template_uri = 'http://example.org/sbol3/template1'
        ss = SampleSet(identity=ss_uri, template=template_uri)
        doc.add(ss)

        doc2 = sbol3.Document()
        rdf_format = sbol3.TURTLE
        output = doc.write_string(rdf_format)
        doc2.read_string(output, rdf_format)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestCombinatorialDerivation(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        template = 'https://github.com/synbiodex/pysbol3/component'
        cd1 = sbol3.CombinatorialDerivation('cd1', template)
        self.assertEqual(template, cd1.template)
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        cd2 = sbol3.CombinatorialDerivation('cd2', comp1)
        self.assertEqual(comp1.identity, cd2.template)

    def test_invalid_strategy(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        # Test with invalid strategy
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        cd1 = sbol3.CombinatorialDerivation('cd1', comp1)
        cd1.strategy = sbol3.SBOL_INLINE
        report = cd1.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))

    def test_round_trip(self):
        # See https://github.com/SynBioDex/pySBOL3/issues/156
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        cd1 = sbol3.CombinatorialDerivation('cd1', comp1)
        self.assertEqual(comp1.identity, cd1.template)
        doc1 = sbol3.Document()
        doc1.add(comp1)
        doc1.add(cd1)
        doc2 = sbol3.Document()
        doc2.read_string(doc1.write_string(sbol3.SORTED_NTRIPLES),
                         sbol3.SORTED_NTRIPLES)
        comp2 = doc2.find(comp1.identity)
        self.assertIsInstance(comp2, sbol3.Component)
        cd2 = doc2.find(cd1.identity)
        self.assertIsInstance(cd2, sbol3.CombinatorialDerivation)


if __name__ == '__main__':
    unittest.main()



import abc
from typing import Union

from . import *

int_property = Union[IntProperty, int]


class Location(Identified, abc.ABC):
    """The Location class is used to represent the location of Features
    within Sequences. This class is extended by the Range, Cut, and
    EntireSequence classes

    """

    def __init__(self, sequence: Union[Sequence, str],
                 identity: str, type_uri: str,
                 *, orientation: str = None,
                 order: int = None) -> None:
        super().__init__(identity, type_uri)
        self.orientation = URIProperty(self, SBOL_ORIENTATION, 0, 1,
                                       initial_value=orientation)
        self.order = IntProperty(self, SBOL_ORDER, 0, 1,
                                 initial_value=order)
        self.sequence = ReferencedObject(self, SBOL_SEQUENCES, 1, 1,
                                         initial_value=sequence)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if not self.sequence:
            message = f'Location {self.identity} does not have a sequence'
            report.addError(self.identity, None, message)
        return report


class Range(Location):
    """A Range object specifies a region via discrete, inclusive start and
    end positions that correspond to indices for characters in the
    elements String of a Sequence.

    Note that the index of the first location is 1, as is typical
    practice in biology, rather than 0, as is typical practice in
    computer science.

    """

    def __init__(self, sequence: Union[Sequence, str], start: int, end: int,
                 *, orientation: str = None,
                 order: int = None,
                 identity: str = None, type_uri: str = SBOL_RANGE) -> None:
        super().__init__(sequence=sequence, orientation=orientation,
                         order=order, identity=identity,
                         type_uri=type_uri)
        self.start: int_property = IntProperty(self, SBOL_START, 1, 1,
                                               initial_value=start)
        self.end: int_property = IntProperty(self, SBOL_END, 1, 1,
                                             initial_value=end)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if self.start < 1:
            message = 'Range.start must be greater than 0'
            report.addError(self.identity, 'sbol3-11401', message)
        # TODO: start must also be less than or equal to len(sequence)
        if self.end < 1:
            message = 'Range.end must be greater than 0'
            report.addError(self.identity, 'sbol3-11402', message)
        # TODO: end must also be less than or equal to len(sequence)
        if self.end < self.start:
            message = 'Range.end must be >= start'
            report.addError(self.identity, 'sbol3-11403', message)
        return report


def build_range(identity: str, type_uri: str = SBOL_RANGE):
    """Used by Document to construct a Range when reading an SBOL file.
    """
    start = 1
    end = 1
    obj = Range(PYSBOL3_MISSING, start, end, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SEQUENCES] = []
    obj._properties[SBOL_START] = []
    obj._properties[SBOL_END] = []
    return obj


Document.register_builder(SBOL_RANGE, build_range)


class Cut(Location):
    """The Cut class has been introduced to enable the specification of a
    region between two discrete positions. This specification is
    accomplished using the at property, which specifies a discrete
    position that corresponds to the index of a character in the
    elements String of a Sequence (except in the case when at is equal
    to zero).

    """

    def __init__(self, sequence: Union[Sequence, str], at: int,
                 *, orientation: str = None,
                 order: int = None,
                 identity: str = None, type_uri: str = SBOL_CUT) -> None:
        super().__init__(sequence=sequence, orientation=orientation,
                         order=order, identity=identity,
                         type_uri=type_uri)
        self.at: int_property = IntProperty(self, SBOL_START, 1, 1,
                                            initial_value=at)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        if self.at < 0:
            message = 'Cut property "at" must be >= 0'
            report.addError(self.identity, None, message)
        return report


def build_cut(identity: str, type_uri: str = SBOL_CUT):
    """Used by Document to construct a Cut when reading an SBOL file.
    """
    at = 0
    obj = Cut(PYSBOL3_MISSING, at, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SEQUENCES] = []
    obj._properties[SBOL_START] = []
    return obj


Document.register_builder(SBOL_CUT, build_cut)


class EntireSequence(Location):
    """The EntireSequence class does not have any additional
    properties. Use of this class indicates that the linked Sequence
    describes the entirety of the Component or Feature parent of this
    Location object.

    """

    def __init__(self, sequence: Union[Sequence, str],
                 *, orientation: str = None,
                 order: int = None,
                 identity: str = None,
                 type_uri: str = SBOL_ENTIRE_SEQUENCE) -> None:
        super().__init__(sequence=sequence, orientation=orientation,
                         order=order, identity=identity,
                         type_uri=type_uri)


def build_entire_sequence(identity: str, type_uri: str = SBOL_ENTIRE_SEQUENCE):
    """Used by Document to construct an EntireSequence when reading an SBOL file.
    """
    obj = EntireSequence(PYSBOL3_MISSING, identity=identity, type_uri=type_uri)
    obj._properties[SBOL_SEQUENCES] = []
    return obj


Document.register_builder(SBOL_ENTIRE_SEQUENCE, build_entire_sequence)



import datetime
import math
from typing import Union, List

from . import *


class Usage(CustomIdentified):
    """How different entities are used in an prov:Activity is specified
    with the prov:Usage class, which is linked from an prov:Activity
    through the prov:Usage relationship. A prov:Usage is then linked
    to an prov:Entity through the prov:entity property URI and the
    prov:hadRole property species how the prov:Entity is used. When
    the prov:wasDerivedFrom property is used together with the full
    provenance described here, the entity pointed at by the
    prov:wasDerivedFrom property MUST be included in a prov:Usage.

    """

    def __init__(self, entity: str,
                 *, roles: str = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = PROV_USAGE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.entity = URIProperty(self, PROV_ENTITY, 1, 1,
                                  initial_value=entity)
        self.roles = URIProperty(self, PROV_ROLES, 0, math.inf,
                                 initial_value=roles)


def build_usage(identity: str, *, type_uri: str = PROV_USAGE) -> SBOLObject:
    obj = Usage(entity=PYSBOL3_MISSING, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[PROV_ENTITY] = []
    return obj


Document.register_builder(PROV_USAGE, build_usage)


class Agent(CustomTopLevel):
    """Examples of agents are a person, organization, or software
    tool. These agents should be annotated with additional
    information, such as software version, needed to be able to run
    the same prov:Activity again.

    """

    def __init__(self, identity: str,
                 *, attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = PROV_AGENT) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


Document.register_builder(PROV_AGENT, Agent)


class Plan(CustomTopLevel):
    """The prov:Plan entity can be used as a place holder to describe the
    steps (for example scripts or lab protocols) taken when an
    prov:Agent is used in a particular prov:Activity.

    """

    def __init__(self, identity: str,
                 *, attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = PROV_PLAN) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


Document.register_builder(PROV_PLAN, Plan)


class Association(CustomIdentified):
    """An prov:Association is linked to an prov:Agent through the
    prov:agent relationship. The prov:Association includes the
    prov:hadRole property to qualify the role of the prov:Agent in the
    prov:Activity.

    """

    def __init__(self, agent: Union[str, Identified],
                 *, roles: str = None, plan: Union[Identified, str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = PROV_ASSOCIATION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.roles = URIProperty(self, PROV_ROLES, 0, math.inf,
                                 initial_value=roles)
        self.plan = ReferencedObject(self, PROV_PLANS, 0, 1,
                                     initial_value=plan)
        self.agent = ReferencedObject(self, PROV_AGENTS, 1, 1,
                                      initial_value=agent)


def build_association(identity: str,
                      *, type_uri: str = PROV_USAGE) -> SBOLObject:
    obj = Association(agent=PYSBOL3_MISSING, identity=identity,
                      type_uri=type_uri)
    # Remove the dummy values
    obj._properties[PROV_AGENTS] = []
    return obj


Document.register_builder(PROV_ASSOCIATION, build_association)


class Activity(CustomTopLevel):
    """A generated prov:Entity is linked through a prov:wasGeneratedBy
    relationship to an prov:Activity, which is used to describe how
    different prov:Agents and other entities were used. An
    prov:Activity is linked through a prov:qualifiedAssociation to
    prov:Associations, to describe the role of agents, and is linked
    through prov:qualifiedUsage to prov:Usages to describe the role of
    other entities used as part of the activity. Moreover, each
    prov:Activity includes optional prov:startedAtTime and
    prov:endedAtTime properties. When using prov:Activity to capture
    how an entity was derived, it is expected that any additional
    information needed will be attached as annotations. This may
    include software settings or textual notes. Activities can also be
    linked together using the prov:wasInformedBy relationship to
    provide dependency without explicitly specifying start and end
    times.

    """

    def __init__(self, identity: str,
                 *, types: List[str] = None,
                 start_time: Union[str, datetime.datetime] = None,
                 end_time: Union[str, datetime.datetime] = None,
                 usage: List[Identified] = None,
                 association: List[Identified] = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = PROV_ACTIVITY) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.types = URIProperty(self, SBOL_TYPE, 0, math.inf,
                                 initial_value=types)
        self.start_time = DateTimeProperty(self, PROV_STARTED_AT_TIME, 0, 1,
                                           initial_value=start_time)
        self.end_time = DateTimeProperty(self, PROV_ENDED_AT_TIME, 0, 1,
                                         initial_value=end_time)
        self.usage = OwnedObject(self, PROV_QUALIFIED_USAGE, 0, math.inf,
                                 initial_value=usage,
                                 type_constraint=Usage)
        self.association = OwnedObject(self, PROV_QUALIFIED_ASSOCIATION,
                                       0, math.inf,
                                       initial_value=association,
                                       type_constraint=Association)


Document.register_builder(PROV_ACTIVITY, Activity)



import abc
from typing import Union, List

from . import *

from .om_unit import Unit


class CompoundUnit(Unit, abc.ABC):
    """As adopted by SBOL, om:CompoundUnit is an abstract class that is
    extended by other classes to describe units of measure that can be
    represented as combinations of multiple other units of measure.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 type_uri: str,
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None) -> None:
        super().__init__(symbol=symbol, label=label,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)


class UnitMultiplication(CompoundUnit):
    """The purpose of the om:UnitMultiplication class is to describe a
    unit of measure that is the multiplication of two other units of
    measure.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 term1: Union[Unit, str], term2: Union[Unit, str],
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_UNIT_MULTIPLICATION) -> None:
        super().__init__(symbol=symbol, label=label,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.term1 = ReferencedObject(self, OM_HAS_TERM1, 1, 1,
                                      initial_value=term1)
        self.term2 = ReferencedObject(self, OM_HAS_TERM2, 1, 1,
                                      initial_value=term2)


def build_unit_multiplication(identity: str,
                              *, type_uri: str = OM_UNIT_MULTIPLICATION) -> SBOLObject:
    obj = UnitMultiplication(identity=identity, type_uri=type_uri,
                             symbol=PYSBOL3_MISSING, label=PYSBOL3_MISSING,
                             term1=PYSBOL3_MISSING, term2=PYSBOL3_MISSING)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_TERM1] = []
    obj._properties[OM_HAS_TERM2] = []
    return obj


Document.register_builder(OM_UNIT_MULTIPLICATION, build_unit_multiplication)


class UnitDivision(CompoundUnit):
    """The purpose of the om:UnitDivision class is to describe a unit of
    measure that is the division of one unit of measure by another.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 numerator: Union[Unit, str], denominator: Union[Unit, str],
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_UNIT_DIVISION) -> None:
        super().__init__(symbol=symbol, label=label,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.numerator = ReferencedObject(self, OM_HAS_NUMERATOR, 1, 1,
                                          initial_value=numerator)
        self.denominator = ReferencedObject(self, OM_HAS_DENOMINATOR, 1, 1,
                                            initial_value=denominator)


def build_unit_division(identity: str,
                        *, type_uri: str = OM_UNIT_DIVISION) -> SBOLObject:
    obj = UnitDivision(identity=identity, type_uri=type_uri,
                       symbol=PYSBOL3_MISSING, label=PYSBOL3_MISSING,
                       numerator=PYSBOL3_MISSING, denominator=PYSBOL3_MISSING)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_NUMERATOR] = []
    obj._properties[OM_HAS_DENOMINATOR] = []
    return obj


Document.register_builder(OM_UNIT_DIVISION, build_unit_division)


class UnitExponentiation(CompoundUnit):
    """The purpose of the om:UnitExponentiation class is to describe a
    unit of measure that is raised to an integer power.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 base: Union[Unit, str], exponent: int,
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_UNIT_EXPONENTIATION) -> None:
        super().__init__(symbol=symbol, label=label,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.base = ReferencedObject(self, OM_HAS_BASE, 1, 1,
                                     initial_value=base)
        self.exponent = IntProperty(self, OM_HAS_EXPONENT, 1, 1,
                                    initial_value=exponent)


def build_unit_exponentiation(identity: str,
                              *, type_uri: str = OM_UNIT_EXPONENTIATION) -> SBOLObject:
    obj = UnitExponentiation(identity=identity, type_uri=type_uri,
                             symbol=PYSBOL3_MISSING, label=PYSBOL3_MISSING,
                             base=PYSBOL3_MISSING, exponent=1)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_BASE] = []
    obj._properties[OM_HAS_EXPONENT] = []
    return obj


Document.register_builder(OM_UNIT_EXPONENTIATION, build_unit_exponentiation)



import abc
import math
from typing import Union, List

from . import *
from .om_prefix import Prefix


class Unit(CustomTopLevel, abc.ABC):
    """As adopted by SBOL, om:Unit is an abstract class that is extended
    by other classes to describe units of measure using a shared set
    of properties.

    See Appendix A Section A.2 of the SBOL 3.0 specificiation.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 type_uri: str,
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.symbol = TextProperty(self, OM_SYMBOL, 1, 1,
                                   initial_value=symbol)
        self.alternative_symbols = TextProperty(self, OM_ALTERNATIVE_SYMBOL,
                                                0, math.inf,
                                                initial_value=alternative_symbols)
        self.label = TextProperty(self, OM_LABEL, 1, 1,
                                  initial_value=label)
        self.alternative_labels = TextProperty(self, OM_ALTERNATIVE_LABEL,
                                               0, math.inf,
                                               initial_value=alternative_labels)
        self.comment = TextProperty(self, OM_COMMENT, 0, 1,
                                    initial_value=comment)
        self.long_comment = TextProperty(self, OM_LONG_COMMENT, 0, 1,
                                         initial_value=long_comment)


class Measure(CustomIdentified):
    """The purpose of the om:Measure class is to link a numerical value to
    a om:Unit.

    """

    def __init__(self, value: float, unit: str,
                 *, types: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 identity: str = None,
                 type_uri: str = OM_MEASURE) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        self.value = FloatProperty(self, OM_HAS_NUMERICAL_VALUE, 1, 1,
                                   initial_value=value)
        self.types = URIProperty(self, SBOL_TYPE, 0, math.inf,
                                 initial_value=types)
        self.unit = URIProperty(self, OM_HAS_UNIT, 1, 1,
                                initial_value=unit)


def build_measure(identity: str, *, type_uri: str = OM_MEASURE) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Measure(1.0, missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_HAS_NUMERICAL_VALUE] = []
    obj._properties[OM_HAS_UNIT] = []
    return obj


Document.register_builder(OM_MEASURE, build_measure)


class SingularUnit(Unit):
    """The purpose of the om:SingularUnit class is to describe a unit of
    measure that is not explicitly represented as a combination of
    multiple units, but could be equivalent to such a
    representation. For example, a joule is considered to be a
    om:SingularUnit, but it is equivalent to the multiplication of a
    newton and a meter.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 *, unit: str = None, factor: float = None,
                 alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_SINGULAR_UNIT) -> None:
        super().__init__(symbol=symbol, label=label,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.unit = ReferencedObject(self, OM_HAS_UNIT, 0, 1,
                                     initial_value=unit)
        self.factor = FloatProperty(self, OM_HAS_FACTOR, 0, 1,
                                    initial_value=factor)


def build_singular_unit(identity: str,
                        *, type_uri: str = OM_SINGULAR_UNIT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = SingularUnit(symbol=missing, label=missing,
                       identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    return obj


Document.register_builder(OM_SINGULAR_UNIT, build_singular_unit)


class PrefixedUnit(Unit):
    """The purpose of the om:PrefixedUnit class is to describe a unit of
    measure that is the multiplication of another unit of measure and
    a factor represented by a standard prefix such as “milli,”
    “centi,” “kilo,” etc.

    """

    def __init__(self, identity: str, symbol: str, label: str,
                 unit: Union[Unit, str],
                 prefix: Union[Prefix, str],
                 *, alternative_symbols: List[str] = None,
                 alternative_labels: List[str] = None,
                 comment: str = None,
                 long_comment: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = OM_PREFIXED_UNIT) -> None:
        super().__init__(symbol=symbol, label=label,
                         identity=identity, type_uri=type_uri,
                         alternative_symbols=alternative_symbols,
                         alternative_labels=alternative_labels,
                         comment=comment, long_comment=long_comment,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.unit = ReferencedObject(self, OM_HAS_UNIT, 1, 1,
                                     initial_value=unit)
        self.prefix = ReferencedObject(self, OM_HAS_PREFIX, 1, 1,
                                       initial_value=prefix)


def build_prefixed_unit(identity: str,
                        *, type_uri: str = OM_PREFIXED_UNIT) -> SBOLObject:
    obj = PrefixedUnit(identity=identity, type_uri=type_uri,
                       symbol=PYSBOL3_MISSING, label=PYSBOL3_MISSING,
                       unit=PYSBOL3_MISSING, prefix=PYSBOL3_MISSING)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_UNIT] = []
    obj._properties[OM_HAS_PREFIX] = []
    return obj


Document.register_builder(OM_PREFIXED_UNIT, build_prefixed_unit)



import unittest

import sbol3


class TestSequenceFeature(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        es1 = sbol3.EntireSequence(sbol3.PYSBOL3_MISSING)
        locations = [es1]
        sf = sbol3.SequenceFeature(locations)
        self.assertEqual(locations, sf.locations)

    def test_validation(self):
        locations = []
        sf = sbol3.SequenceFeature(locations)
        report = sf.validate()
        self.assertIsNotNone(report)
        self.assertEqual(1, len(report.errors))

    def test_recursive_validation(self):
        # Test that the owned object, in this case the Range,
        # is also validated when the SequenceFeature is validated.
        seq_uri = 'https://github.com/synbiodex/pysbol3/sequence'
        start = 10
        end = 1
        r = sbol3.Range(seq_uri, start, end)
        report = r.validate()
        # end < start is a validation error
        self.assertEqual(1, len(report.errors))
        sf = sbol3.SequenceFeature([r])
        report = sf.validate()
        # We should find the validation issue in the owned
        # object (the range).
        self.assertEqual(1, len(report.errors))


if __name__ == '__main__':
    unittest.main()



from typing import List

from . import *


class Attachment(TopLevel):
    """The purpose of the Attachment class is to serve as a general
    container for data files, especially experimental data files. It
    provides a means for linking files and metadata to SBOL designs.

    """

    def __init__(self, identity: str, source: str,
                 *, format: str = None, size: int = None,
                 hash: str = None, hash_algorithm: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_ATTACHMENT):
        super().__init__(identity=identity, type_uri=type_uri,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.source = URIProperty(self, SBOL_SOURCE, 1, 1,
                                  initial_value=source)
        self.format = URIProperty(self, SBOL_FORMAT, 0, 1,
                                  initial_value=format)
        self.size = IntProperty(self, SBOL_SIZE, 0, 1,
                                initial_value=size)
        self.hash = TextProperty(self, SBOL_HASH, 0, 1,
                                 initial_value=hash)
        self.hash_algorithm = TextProperty(self, SBOL_HASH_ALGORITHM, 0, 1,
                                           initial_value=hash_algorithm)

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        # An Attachment must have 1 source
        if self.source is None:
            message = f'Attachment {self.identity} must have a source'
            report.addError(self.identity, None, message)
        return report


def build_attachment(identity: str,
                     *, type_uri: str = SBOL_COMPONENT) -> SBOLObject:
    obj = Attachment(identity=identity, source=PYSBOL3_MISSING,
                     type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SOURCE] = []
    return obj


Document.register_builder(SBOL_ATTACHMENT, build_attachment)
